import inspect
from . import util as util
from _typeshed import Incomplete
from collections.abc import Generator

EMPTY: Incomplete

def null_context(owner) -> Generator[Incomplete, None, None]: ...

class NeverRaisedException(BaseException): ...

class notify:
    @classmethod
    def clear(cls) -> None: ...
    @classmethod
    def hold_owner_notifications(cls, *owners) -> None: ...
    @classmethod
    def allow_owner_notifications(cls, *owners) -> None: ...
    @classmethod
    def return_event(cls, owner, returned: dict): ...
    @classmethod
    def call_event(cls, owner, parameters: dict): ...
    @classmethod
    def call_iteration_event(
        cls, owner, index: int, step_name: str = ..., total_count: int = ...
    ): ...
    @classmethod
    def observe_returns(cls, handler) -> None: ...
    @classmethod
    def observe_calls(cls, handler) -> None: ...
    @classmethod
    def observe_call_iteration(cls, handler) -> None: ...
    @classmethod
    def unobserve_returns(cls, handler) -> None: ...
    @classmethod
    def unobserve_calls(cls, handler) -> None: ...
    @classmethod
    def unobserve_call_iteration(cls, handler) -> None: ...

class CallSignatureTemplate:
    target: Incomplete

    def __init__(self, target) -> None: ...
    def get_target(self, owner): ...
    def get_keyword_parameters(self, owner, skip_names): ...

class MethodTaggerDataclass:
    pending: Incomplete

    def __call__(self, func): ...

class rack_input_table(MethodTaggerDataclass):
    table_path: str

    def __init__(self, table_path) -> None: ...

class rack_kwargs_template(MethodTaggerDataclass):
    template: callable

    def __init__(self, template) -> None: ...

class rack_kwargs_skip(MethodTaggerDataclass):
    skip: list

    def __init__(self, *arg_names) -> None: ...

class RackMethod(util.Ownable):
    tags: Incomplete
    __doc__: Incomplete
    __name__: Incomplete
    __qualname__: Incomplete

    def __init__(self, owner, name: str, kwdefaults: dict = ...) -> None: ...
    def iterate_from_csv(self, path) -> Generator[Incomplete, None, None]: ...
    debug: Incomplete

    @classmethod
    def from_method(self, method): ...
    def __copy__(self): ...
    def __deepcopy__(self, memo: Incomplete | None = ...): ...
    def __owner_subclass__(self, owner_cls) -> None: ...
    def set_kwdefault(self, name, new_default) -> None: ...
    def extended_signature(self, name_map=...): ...
    def extended_arguments(self, name_map=...): ...
    def call_by_extended_argnames(self, *args, **kws): ...
    def __call__(self, *args, **kws): ...

class BoundSequence(util.Ownable):
    cleanup_func: Incomplete
    exception_allowlist = NeverRaisedException

    def __call__(self, **kwargs): ...
    @classmethod
    def to_template(cls, path) -> None: ...
    def iterate_from_csv(self, path) -> Generator[Incomplete, None, None]: ...

class OwnerContextAdapter:
    def __init__(self, owner) -> None: ...
    def __enter__(self) -> None: ...
    def __exit__(self, *exc_info) -> None: ...

def flatten_nested_owner_contexts(top) -> dict: ...
def package_owned_contexts(top): ...
def owner_getattr_chains(owner): ...

class Owner:
    def __init_subclass__(cls, entry_order: list = ...): ...
    def __init__(self, **update_ownables) -> None: ...
    def __propagate_ownership__(self) -> None: ...
    def __setattr__(self, key, obj) -> None: ...
    def close(self) -> None: ...
    def open(self) -> None: ...
    @property
    def __enter__(self): ...
    @property
    def __exit__(self): ...

def recursive_devices(top: Owner): ...
def override_empty(a, b, param_name, field): ...
def update_parameter_dict(dest: dict, signature: inspect.Signature): ...
def attr_chain_to_method(root_obj, chain): ...
def standardize_spec_step(sequence): ...

class Sequence(util.Ownable):
    access_spec: Incomplete
    cleanup_func: Incomplete
    exception_allowlist = NeverRaisedException
    spec: Incomplete
    tags: Incomplete

    def __init__(
        self, *specification, shared_names=..., input_table: Incomplete | None = ...
    ) -> None: ...
    def return_on_exceptions(
        self, exception_or_exceptions, cleanup_func: Incomplete | None = ...
    ) -> None: ...
    def __owner_subclass__(self, owner_cls): ...
    last_spec: Incomplete

    def __owner_init__(self, owner) -> None: ...

class RackMeta(type):
    def __enter__(cls) -> None: ...
    def __exit__(cls, *exc_info) -> None: ...

class Rack(Owner, util.Ownable, metaclass=RackMeta):
    def __init__(): ...
    def __init_subclass__(cls, entry_order=...) -> None: ...
    def __deepcopy__(self, memo: Incomplete | None = ...): ...
    def __owner_init__(self, owner) -> None: ...
    def __getattribute__(self, item): ...
    def __getitem__(self, item): ...
    def __len__(self) -> int: ...
    def __iter__(self): ...

class _use_module_path:
    path: Incomplete

    def __init__(self, path) -> None: ...
    def __enter__(self) -> None: ...
    def __exit__(self, exc_type, exc_value, traceback) -> None: ...

def import_as_rack(
    import_string: str,
    *,
    cls_name: str = ...,
    append_path: list = ...,
    base_cls: type = ...,
    replace_attrs: list = ...,
): ...
def find_owned_rack_by_type(
    parent_rack: Rack, target_type: Rack, include_parent: bool = ...
): ...



from . import util, value as value
from ._traits import HasTraits
from _typeshed import Incomplete

def list_devices(depth: int = ...): ...

class DisconnectedBackend:
    name: Incomplete

    def __init__(self, dev) -> None: ...
    def __getattr__(self, key) -> None: ...
    def __copy__(self, memo: Incomplete | None = ...): ...
    str: Incomplete
    __deepcopy__ = __copy__

class Device(HasTraits, util.Ownable):
    def __init__(self, resource: str = "str"): ...
    resource: Incomplete
    concurrency: Incomplete
    backend: Incomplete

    def open(self) -> None: ...
    def close(self) -> None: ...
    __children__: Incomplete

    @classmethod
    def __init_subclass__(cls) -> None: ...
    @classmethod
    def __update_signature__(cls) -> None: ...
    def __open_wrapper__(self) -> None: ...
    def __close_wrapper__(self) -> None: ...
    def __enter__(self): ...
    def __exit__(self, type_, value, traceback) -> None: ...
    def __del__(self) -> None: ...
    def isopen(self): ...

def trait_info(device: Device, name: str) -> dict: ...



import labbench as lb
import pandas as pd
import numpy as np


@lb.property.visa_keying(
    # the default SCPI query and write formats
    query_fmt="{key}?",
    write_fmt="{key} {value}",
    # map python True and False values to these SCPI strings
    remap={True: "ON", False: "OFF"},
)
@lb.adjusted(
    # set the identity_pattern
    "identity_pattern",
    default=r"Power Sensor model \#1234",
)
class PowerSensor(lb.VISADevice):
    RATES = "NORM", "DOUB", "FAST"

    # SCPI string keys and bounds on the parameter values,
    # taken from the instrument programming manual
    initiate_continuous = lb.property.bool(
        key="INIT:CONT", help="trigger continuously if True"
    )
    trigger_count = lb.property.int(
        key="TRIG:COUN", min=1, max=200, help="acquisition count", label="samples"
    )
    measurement_rate = lb.property.str(
        key="SENS:MRAT",
        only=RATES,
        case=False,
    )
    sweep_aperture = lb.property.float(
        key="SWE:APER", min=20e-6, max=200e-3, help="measurement duration", label="s"
    )
    frequency = lb.property.float(
        key="SENS:FREQ",
        min=10e6,
        max=18e9,
        step=1e-3,
        help="calibration frequency",
        label="Hz",
    )

    def preset(self):
        """revert to instrument default presets"""
        self.write("SYST:PRES")

    def fetch(self):
        """acquire measurements as configured"""
        response = self.query("FETC?")

        if self.trigger_count == 1:
            return float(response)
        else:
            return pd.Series([float(s) for s in response.split(",")], name="spectrum")

    def trigger(self):
        return self.write("TRIG")


@lb.property.visa_keying(remap={True: "ON", False: "OFF"})
@lb.adjusted("identity_pattern", default=r"Spectrum analyzer model \#1234")
class SpectrumAnalyzer(lb.VISADevice):
    center_frequency = lb.property.float(
        key="SENS:FREQ",
        min=10e6,
        max=18e9,
        step=1e-3,
        help="input signal center frequency",
        label="Hz",
    )
    resolution_bandwidth = lb.property.float(
        key="SENS:BW",
        min=1,
        max=40e6,
        step=1e-3,
        help="resolution bandwidth",
        label="Hz",
    )

    def load_state(self, remote_filename: str):
        """revert to instrument preset state"""
        self.write(f"LOAD '{remote_filename}'")

    def fetch(self):
        """acquire measurements as configured"""
        response = self.query("FETC?")

        series = pd.Series([float(s) for s in response.split(",")], name="spectrum")
        series.index = pd.Index(
            self.center_frequency + np.linspace(-5e6, 5e6, len(series)),
            name="frequency",
        )

        return series

    def trigger(self):
        return self.write("TRIG")


@lb.property.visa_keying(remap={True: "YES", False: "NO"})
@lb.adjusted("identity_pattern", default=r"Signal generator model \#1234")
class SignalGenerator(lb.VISADevice):
    output_enabled = lb.property.bool(
        key="OUT:ENABL", help="when True, output an RF tone"
    )
    center_frequency = lb.property.float(
        key="SENS:FREQ",
        min=10e6,
        max=18e9,
        step=1e-3,
        help="input signal center frequency",
        label="Hz",
    )
    mode = lb.property.str(key="MODE", only=["sweep", "tone", "iq"], case=False)

    def trigger(self):
        """revert to instrument preset state"""
        self.write("TRIG")



import labbench as lb


class EmulatedVISAPropertyAdapter(lb.VISAPropertyAdapter):
    def get(self, device, key, trait):
        import numpy as np

        if trait.type is bool:
            if len(self.remap) > 0:
                return np.random.choice(self.value_map.values())
            else:
                return np.random.choice(("TRUE", "FALSE"))

        elif trait.type is str:
            return "text"

        elif trait.type is float:
            return str(np.random.uniform(low=trait.min, high=trait.max))

        else:
            raise TypeError("No emulated values implemented for trait {repr(trait)}")

    def set(self, device, key, value, trait):
        pass


@EmulatedVISAPropertyAdapter
class EmulatedVISADevice(lb.Device):
    """Act as a VISA device without dispatching any visa keys"""

    # Settings
    read_termination = lb.value.str("\n", help="end-of-receive termination character")
    write_termination = lb.value.str("\n", help="end-of-transmit termination character")

    # States
    @lb.property.str(key="*IDN", sets=False, cache=True)
    def identity(self):
        """identity string reported by the instrument"""
        return self.__class__.__qualname__

    @lb.property.str(key="*OPT", sets=False, cache=True)
    def options(self):
        """options reported by the instrument"""

        param_strs = (f"{s}={getattr(self,s)}" for s in self._value_attrs)
        return ",".join(param_strs)

    @lb.property.dict(key="*STB", sets=False)
    def status_byte(self):
        """VISA status byte reported by the instrument"""
        return {
            "error queue not empty": False,
            "questionable state": False,
            "message available": False,
            "event status flag": False,
            "service request": False,
            "master status summary": False,
            "operating": True,
        }



# This software was developed by employees of the National Institute of
# Standards and Technology (NIST), an agency of the Federal Government.
# Pursuant to title 17 United States Code Section 105, works of NIST employees
# are not subject to copyright protection in the United States and are
# considered to be in the public domain. Permission to freely use, copy,
# modify, and distribute this software and its documentation without fee is
# hereby granted, provided that this notice and disclaimer of warranty appears
# in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' WITHOUT ANY WARRANTY OF ANY KIND, EITHER
# EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, ANY WARRANTY
# THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY IMPLIED WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND FREEDOM FROM
# INFRINGEMENT, AND ANY WARRANTY THAT THE DOCUMENTATION WILL CONFORM TO THE
# SOFTWARE, OR ANY WARRANTY THAT THE SOFTWARE WILL BE ERROR FREE. IN NO EVENT
# SHALL NIST BE LIABLE FOR ANY DAMAGES, INCLUDING, BUT NOT LIMITED TO, DIRECT,
# INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM,
# OR IN ANY WAY CONNECTED WITH THIS SOFTWARE, WHETHER OR NOT BASED UPON
# WARRANTY, CONTRACT, TORT, OR OTHERWISE, WHETHER OR NOT INJURY WAS SUSTAINED
# BY PERSONS OR Decorator OR OTHERWISE, AND WHETHER OR NOT LOSS WAS SUSTAINED
# FROM, OR AROSE OUT OF THE RESULTS OF, OR USE OF, THE SOFTWARE OR SERVICES
# PROVIDED HEREUNDER. Distributions of NIST software should also include
# copyright and licensing statements of any third-party software that are
# legally bundled with the code in compliance with the conditions of those
# licenses.

import unittest
import importlib
import sys
import labbench as lb

lb.util._force_full_traceback(True)

from test_value import *
from test_property import *

# from test_db import *
# from test_concurrently import *

# from test_doc import *

if __name__ == "__main__":
    lb.show_messages("warning")
    unittest.main()



from . import _traits
from _typeshed import Incomplete

class message_keying(_traits.PropertyKeyingBase):
    query_fmt: Incomplete
    write_fmt: Incomplete
    write_func: Incomplete
    query_func: Incomplete
    value_map: Incomplete
    message_map: Incomplete

    def __init__(
        self,
        query_fmt: Incomplete | None = ...,
        write_fmt: Incomplete | None = ...,
        write_func: Incomplete | None = ...,
        query_func: Incomplete | None = ...,
        remap=...,
    ) -> None: ...
    def get(
        self, device: _traits.HasTraits, scpi_key: str, trait: Incomplete | None = ...
    ): ...
    def set(
        self,
        device: _traits.HasTraits,
        scpi_key: str,
        value,
        trait: Incomplete | None = ...,
    ): ...

class visa_keying(message_keying):
    def __init__(
        self,
        query_fmt: str = ...,
        write_fmt: str = ...,
        write_func: str = ...,
        query_func: str = ...,
        remap=...,
    ) -> None: ...

class bool(_traits.Bool):
    def __init__(
        key=None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
    ): ...
    ...

class float(_traits.Float):
    def __init__(
        key=None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = True,
        min: str = None,
        max: str = None,
        path_trait: str = None,
        index_lookup_trait: str = None,
        table_index_column: str = None,
        step: str = None,
    ): ...
    ...

class int(_traits.Int):
    def __init__(
        key=None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = True,
        min: str = None,
        max: str = None,
        path_trait: str = None,
        index_lookup_trait: str = None,
        table_index_column: str = None,
    ): ...
    ...

class complex(_traits.Complex):
    def __init__(
        key=None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
    ): ...
    ...

class str(_traits.Unicode):
    def __init__(
        key=None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
        case: str = True,
    ): ...
    ...

class bytes(_traits.Bytes):
    def __init__(
        key=None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
        case: str = True,
    ): ...
    ...

class list(_traits.List):
    def __init__(
        key=None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
    ): ...
    ...

class tuple(_traits.Tuple):
    def __init__(
        key=None,
        help: str = "",
        label: str = "",
        sets: str = False,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
    ): ...
    ...

class dict(_traits.Dict):
    def __init__(
        key=None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
    ): ...
    ...

class Path(_traits.Path):
    def __init__(
        key=None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
        must_exist: str = False,
    ): ...
    ...

class NetworkAddress(_traits.NetworkAddress):
    def __init__(
        key=None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
        case: str = True,
        accept_port: str = True,
    ): ...
    ...



import contextlib
from . import util as util, value as value
from ._device import Device as Device
from ._traits import observe as observe
from _typeshed import Incomplete
from collections.abc import Generator
from typing import Dict

serial: Incomplete
telnetlib: Incomplete

class ShellBackend(Device):
    def __init__(
        self, resource: str = "str", binary_path: str = "NoneType", timeout: str = "int"
    ): ...
    binary_path: Incomplete
    timeout: Incomplete
    backend: Incomplete

    def open(self) -> None: ...
    def run(
        self,
        *argv,
        pipe: bool = ...,
        background: bool = ...,
        check_return: bool = ...,
        check_stderr: bool = ...,
        respawn: bool = ...,
        timeout: Incomplete | None = ...,
    ): ...
    def read_stdout(self, wait_for: int = ...): ...
    def write_stdin(self, text) -> None: ...
    def kill(self) -> None: ...
    def running(self): ...
    def clear_stdout(self) -> None: ...
    def close(self) -> None: ...

class DotNetDevice(Device):
    def __init__(self, resource: str = "str"): ...
    library: Incomplete
    dll_name: Incomplete
    dll: Incomplete

    def open(self) -> None: ...

class LabviewSocketInterface(Device):
    def __init__(
        self,
        resource: str = "str",
        tx_port: str = "int",
        rx_port: str = "int",
        delay: str = "int",
        timeout: str = "int",
        rx_buffer_size: str = "int",
    ): ...
    resource: Incomplete
    tx_port: Incomplete
    rx_port: Incomplete
    delay: Incomplete
    timeout: Incomplete
    rx_buffer_size: Incomplete
    backend: Incomplete

    def open(self) -> None: ...
    def close(self) -> None: ...
    def write(self, msg) -> None: ...
    def read(self, convert_func: Incomplete | None = ...): ...
    def clear(self) -> None: ...

class SerialDevice(Device):
    def __init__(
        self,
        resource: str = "str",
        timeout: str = "int",
        write_termination: str = "bytes",
        baud_rate: str = "int",
        parity: str = "bytes",
        stopbits: str = "int",
        xonxoff: str = "bool",
        rtscts: str = "bool",
        dsrdtr: str = "bool",
    ): ...
    resource: Incomplete
    timeout: Incomplete
    write_termination: Incomplete
    baud_rate: int
    parity: Incomplete
    stopbits: Incomplete
    xonxoff: Incomplete
    rtscts: Incomplete
    dsrdtr: Incomplete
    backend: Incomplete

    def open(self) -> None: ...
    def close(self) -> None: ...
    @classmethod
    def from_hwid(cls, hwid: Incomplete | None = ..., *args, **connection_params): ...
    @staticmethod
    def list_ports(hwid: Incomplete | None = ...): ...

class SerialLoggingDevice(SerialDevice):
    def __init__(
        self,
        resource: str = "str",
        timeout: str = "int",
        write_termination: str = "bytes",
        baud_rate: str = "int",
        parity: str = "bytes",
        stopbits: str = "int",
        xonxoff: str = "bool",
        rtscts: str = "bool",
        dsrdtr: str = "bool",
        poll_rate: str = "float",
        data_format: str = "bytes",
        stop_timeout: str = "float",
        max_queue_size: str = "int",
    ): ...
    poll_rate: Incomplete
    data_format: Incomplete
    stop_timeout: Incomplete
    max_queue_size: Incomplete

    def configure(self) -> None: ...
    def start(self) -> None: ...
    def stop(self) -> None: ...
    def running(self): ...
    def fetch(self): ...
    def clear(self) -> None: ...
    def close(self) -> None: ...

class TelnetDevice(Device):
    def __init__(self, resource: str = "str", timeout: str = "int"): ...
    resource: Incomplete
    timeout: Incomplete
    backend: Incomplete

    def open(self) -> None: ...
    def close(self) -> None: ...

class VISADevice(Device):
    def __init__(
        self,
        resource: str = "str",
        read_termination: str = "str",
        write_termination: str = "str",
        open_timeout: str = "NoneType",
        identity_pattern: str = "NoneType",
        timeout: str = "NoneType",
    ): ...
    read_termination: Incomplete
    write_termination: Incomplete
    open_timeout: Incomplete
    identity_pattern: Incomplete
    timeout: Incomplete
    identity: Incomplete

    def status_byte(self): ...
    resource: Incomplete
    backend: Incomplete

    def open(self) -> None: ...
    def close(self) -> None: ...
    def write(self, msg: str): ...
    def query(
        self, msg: str, timeout: Incomplete | None = ..., remap: bool = ...
    ) -> str: ...
    def query_ascii_values(
        self,
        msg: str,
        type_,
        separator: str = ...,
        container=...,
        delay: Incomplete | None = ...,
        timeout: Incomplete | None = ...,
    ): ...
    def wait(self) -> None: ...
    def preset(self) -> None: ...
    def overlap_and_block(
        self, timeout: Incomplete | None = ..., quiet: bool = ...
    ) -> Generator[None, None, None,]: ...

    class suppress_timeout(contextlib.suppress):
        def __exit__(self, exctype, excinst, exctb): ...

def visa_list_resources(resourcemanager: str = ...): ...
def visa_default_resource_manager(name: Incomplete | None = ...) -> None: ...
def visa_list_identities(skip_interfaces=..., **device_kws) -> Dict[str, str]: ...

class Win32ComDevice(Device):
    def __init__(self, resource: str = "str"): ...
    com_object: Incomplete
    backend: Incomplete

    def open(self): ...



# This software was developed by employees of the National Institute of
# Standards and Technology (NIST), an agency of the Federal Government.
# Pursuant to title 17 United States Code Section 105, works of NIST employees
# are not subject to copyright protection in the United States and are
# considered to be in the public domain. Permission to freely use, copy,
# modify, and distribute this software and its documentation without fee is
# hereby granted, provided that this notice and disclaimer of warranty appears
# in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' WITHOUT ANY WARRANTY OF ANY KIND, EITHER
# EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, ANY WARRANTY
# THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY IMPLIED WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND FREEDOM FROM
# INFRINGEMENT, AND ANY WARRANTY THAT THE DOCUMENTATION WILL CONFORM TO THE
# SOFTWARE, OR ANY WARRANTY THAT THE SOFTWARE WILL BE ERROR FREE. IN NO EVENT
# SHALL NIST BE LIABLE FOR ANY DAMAGES, INCLUDING, BUT NOT LIMITED TO, DIRECT,
# INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM,
# OR IN ANY WAY CONNECTED WITH THIS SOFTWARE, WHETHER OR NOT BASED UPON
# WARRANTY, CONTRACT, TORT, OR OTHERWISE, WHETHER OR NOT INJURY WAS SUSTAINED
# BY PERSONS OR PROPERTY OR OTHERWISE, AND WHETHER OR NOT LOSS WAS SUSTAINED
# FROM, OR AROSE OUT OF THE RESULTS OF, OR USE OF, THE SOFTWARE OR SERVICES
# PROVIDED HEREUNDER. Distributions of NIST software should also include
# copyright and licensing statements of any third-party software that are
# legally bundled with the code in compliance with the conditions of those
# licenses.

import contextlib
import importlib
import inspect
import logging
import os
import re
import select
import socket
import subprocess as sp
import sys
from collections import OrderedDict
from pathlib import Path
from queue import Empty, Queue
from threading import Event, Thread
from typing import Dict

import psutil
import pyvisa
import pyvisa.errors

from . import property as property_
from . import util, value
from ._device import Device
from ._traits import observe

try:
    serial = util.lazy_import("serial")
    telnetlib = util.lazy_import("telnetlib")
except RuntimeWarning:
    # not executed: help coding tools recognize lazy_imports as imports
    import telnetlib
    import serial


class ShellBackend(Device):
    """Virtual device controlled by a shell command in another process.

    Data can be captured from standard output, and standard error pipes, and
    optionally run as a background thread.

    After opening, `backend` is `None`. On a call to run(background=True), `backend`
    becomes is a subprocess instance. When EOF is reached on the executable's
    stdout, the backend resets to None.

    When `run` is called, the program runs in a subprocess.
    The output piped to the command line standard output is queued in a
    background thread. Call read_stdout() to retreive (and clear) this
    queued stdout.
    """

    binary_path = value.Path(
        default=None, allow_none=True, help="path to the file to run", cache=True
    )

    timeout = value.float(
        default=1,
        min=0,
        help="wait time after close before killing background processes",
        label="s",
        cache=True,
    )

    def open(self):
        """The :meth:`open` method implements opening in the
        :class:`Device` object protocol. Call the
        :meth:`execute` method when open to
        execute the binary.
        """

        def check_state_change(change={}):
            if self.running():
                raise ValueError(
                    "cannot change command line property trait traits during execution"
                )

        if not os.path.exists(self.binary_path):
            raise OSError(
                f'executable does not exist at resource=r"{self.binary_path}"'
            )

        # a Queue for stdout
        self.backend = None

        self._stdout = Queue()
        self._stderr = Queue()

        # Monitor property trait changes
        properties = set(self._value_attrs).difference(dir(ShellBackend))

        observe(self, check_state_change, name=tuple(properties))

    def run(
        self,
        *argv,
        pipe=True,
        background=False,
        check_return=True,
        check_stderr=False,
        respawn=False,
        timeout=None,
    ):
        if pipe and background:
            return self._background_piped(
                *argv,
                check_return=check_return,
                check_stderr=check_stderr,
                timeout=timeout,
                respawn=respawn,
            )

        if respawn:
            raise ValueError("respawn argument requires pipe=True and background=True")

        if pipe and not background:
            return self._run_piped(
                *argv,
                check_return=check_return,
                check_stderr=check_stderr,
                timeout=timeout,
            )
        else:
            if background:
                raise ValueError("background argument requires pipe=True")
            if check_stderr:
                raise ValueError("check_stderr requires pipe=True")

            return self._run_simple(*argv, check_return=check_return, timeout=timeout)

    def _run_simple(self, *argv, check_return=False, timeout=None):
        """Blocking execution of the binary at `self.binary_path`. If check=True, raise an exception
        on non-zero return code.

        Each command line argument in argv is either
        * a string, which is passed to the binary as is, or
        * list-like sequence ('name0', 'name1', ...) that names value traits to insert as flag arguments

        Flag arguments are converted into a sequence of strings by (1) identifying the command line flag
        for each name as `self.flags[name]` (e.g., "-f"), (2) retrieving the value as getattr(self, name), and
        (3) *if* the value is not None, appending the flag to the list of arguments as appropriate.

        Returns:

            None?
        """
        if timeout is None:
            timeout = self.timeout

        return sp.run(self._commandline(argv), check=check_return, timeout=timeout)

    def _run_piped(self, *argv, check_return=False, check_stderr=False, timeout=None):
        """Blocking execution of the binary at `self.binary_path`, with a pipe to collect
        stdout.

        Each command line argument in argv is either
        * a string, which is passed to the binary as is, or
        * list-like sequence ('name0', 'name1', ...) that names value traits to insert as flag arguments

        Flag arguments are converted into a sequence of strings by (1) identifying the command line flag
        for each name as `self.flags[name]` (e.g., "-f"), (2) retrieving the value as getattr(self, name), and
        (3) *if* the value is not None, appending the flag to the list of arguments as appropriate.

        Returns:

            stdout
        """
        if timeout is None:
            timeout = self.timeout

        cmdl = self._commandline(*argv)
        path = cmdl[0]
        try:
            rel = os.path.relpath(path)
            if len(rel) < len(path):
                path = rel
        except ValueError:
            pass

        self._logger.debug(f"shell execute '{repr(' '.join(cmdl))}'")
        cp = sp.run(
            cmdl, timeout=timeout, stdout=sp.PIPE, stderr=sp.PIPE, check=check_return
        )
        ret = cp.stdout

        err = cp.stderr.strip().rstrip().decode()
        if check_stderr and len(err) > 0:
            raise ChildProcessError(err)

        if ret:
            lines = ret.decode().splitlines()
            show_count = min(40, len(lines))
            remaining = max(0, len(lines) - show_count)

            logger_msgs = [f"► {line}" for line in lines[: show_count // 2]]
            if remaining > 0:
                logger_msgs.append(f"…{remaining} more lines")
            for line in lines[-show_count // 2 :]:
                logger_msgs.append(f"► {line}")
            self._logger.debug("\n".join(logger_msgs))
        return ret

    def _background_piped(
        self, *argv, check_return=False, check_stderr=False, respawn=False, timeout=None
    ):
        """Run the executable in the background (returning immediately while
        the executable continues running).

        Once the background process is running,

        * Retrieve standard output from the executable with self.read_stdout

        * Write to standard input self.write_stdin

        * Kill the process with self.kill

        * Check whether the process is running with self.running

        Each command line argument in argv is either
        * a string, which is passed to the binary as is, or
        * list-like sequence ('name0', 'name1', ...) that names value traits to insert as flag arguments

        Flag arguments are converted into a sequence of strings by (1) identifying the command line flag
        for each name as `self.flags[name]` (e.g., "-f"), (2) retrieving the value as getattr(self, name), and
        (3) *if* the value is not None, appending the flag to the list of arguments as appropriate.

        Returns:

            None
        """

        def stdout_to_queue(fd, cmdl):
            """Thread worker to funnel stdout into a queue"""

            pid = self.backend.pid
            q = self._stdout
            for line in iter(fd.readline, ""):
                line = line.decode(errors="replace").replace("\r", "")
                if len(line) > 0:
                    q.put(line)
                else:
                    break
            self.backend = None

            # Respawn (or don't)
            if respawn and not self.__kill:
                self._logger.debug("respawning")
                self._kill_proc_tree(pid)
                spawn(cmdl)
            else:
                self._logger.debug("process ended")

        def stderr_to_exception(fd, cmdl):
            """Thread worker to raise exceptions on standard error output"""
            q = self._stderr
            for line in iter(fd.readline, ""):
                if self.backend is None:
                    break
                line = line.decode(errors="replace").replace("\r", "")
                if len(line) > 0:
                    q.put(line)
                    self._logger.debug(f"stderr {repr(line)}")
                #                    raise Exception(line)
                else:
                    break

        def spawn(cmdl):
            """Execute the binary in the background (nonblocking),
            while funneling its standard output to a queue in a thread.

            Arguments:
                cmd: iterable containing the binary path, then
                        each argument to be passed to the binary.

            Returns:

                None
            """
            if self.running():
                raise Exception("already running")

            si = sp.STARTUPINFO()
            si.dwFlags |= sp.STARTF_USESHOWWINDOW

            proc = sp.Popen(
                list(cmdl),
                stdout=sp.PIPE,
                startupinfo=si,
                creationflags=sp.CREATE_NEW_PROCESS_GROUP,
                stderr=sp.PIPE,
            )

            self.backend = proc
            Thread(target=lambda: stdout_to_queue(proc.stdout, cmdl)).start()
            if check_stderr:
                Thread(target=lambda: stderr_to_exception(proc.stderr, cmdl)).start()

        if not self.isopen:
            raise ConnectionError(
                f"{self} needs to be opened ({self}.open() or in a 'with' block) manage a background process"
            )

        # Generate the commandline and spawn
        cmdl = self._commandline(*argv)
        self._logger.debug(f"background execute: {repr(' '.join(cmdl))}")
        self.__kill = False
        spawn(cmdl)

    def _flags_to_argv(self, flags):
        # find keys in flags that do not exist as value traits
        unsupported = set(flags.keys()).difference(self._value_attrs)
        if len(unsupported) > 1:
            raise KeyError(
                f"flags point to value traits {unsupported} that do not exist in {self}"
            )

        argv = []
        for name, flag_str in flags.items():
            trait = self._traits[name]
            trait_value = getattr(self, name)

            if not isinstance(flag_str, str) and flag_str is not None:
                raise TypeError(
                    f"keys defined in {self} must be str (for a flag) or None (for no flag"
                )

            if trait_value is None:
                continue

            elif trait.type is bool:
                if flag_str is None:
                    # this would require a remap parameter in value traits, which are not supported (should they be?)
                    # (better to use string?)
                    raise ValueError(
                        "cannot map a Bool onto a string argument specified by None mapping"
                    )

                elif trait_value:
                    # trait_value is truey
                    argv += [flag_str]
                    continue

                else:
                    # trait_value is falsey
                    continue

            elif flag_str is None:
                # do not add a flag

                if trait_value is None:
                    # when trait_value is 'None', don't include this flag
                    continue
                else:
                    argv += [str(trait_value)]

            elif isinstance(flag_str, str):
                argv += [flag_str, str(trait_value)]

            else:
                raise ValueError(
                    "unexpected error condition (this should not be possible)"
                )

        return argv

    def _commandline(self, *argv_in):
        """return a new argv list in which dict instances have been replaced by additional
        strings based on the traits in `self`. these dict instances should map {trait_name: cmdline_flag},
        for example dict(force='-f') to map a boolean `self.force` trait value to the -f switch, or
        dict(extra_arg=None) to indicate that `self.extra_arg` will be inserted without a switch if
        `self.extra_arg` is not None.

        Returns:

            tuple of string
        """

        argv = [
            self.binary_path,
        ]

        # Update trait with the flags
        for item in argv_in:
            if isinstance(item, str):
                argv += [item]
            elif isinstance(item, dict):
                argv += self._flags_to_argv(item)
            else:
                raise TypeError(f"command line list item {item} has unsupported type")

        return argv

    def read_stdout(self, wait_for=0):
        """Pop any standard output that has been queued by a background run (see `run`).
        Afterward, the queue is cleared. Starting another background run also clears the queue.

        Returns:

            stdout
        """
        result = ""

        if not self.isopen:
            raise ConnectionError(
                "open the device to read stdout from the background process"
            )

        try:
            n = 0
            while True:
                line = self._stdout.get(wait_for > 0, timeout=self.timeout)
                if isinstance(line, Exception):
                    raise line

                n += 1
                result += line

                if wait_for > 0 and n == wait_for:
                    break
        except Empty:
            pass

        return result

    def write_stdin(self, text):
        """Write characters to stdin if a background process is running. Raises
        Exception if no background process is running.
        """
        try:
            self.backend.stdin.write(text)
        except ConnectionError:
            raise Exception("process not running, could not write no stdin")

    def kill(self):
        """If a process is running in the background, kill it. Sends a console
        warning if no process is running.
        """
        self.__kill = True
        backend = self.backend
        if self.running():
            self._logger.debug(f"killing process {backend.pid}")
            self._kill_proc_tree(backend.pid)

    def running(self):
        """Check whether a background process is running.

        Returns:

            True if running, otherwise False
        """
        # Cache the current running one for a second in case the backend "closes"
        return self.isopen and self.backend is not None and self.backend.poll() is None

    def clear_stdout(self):
        """Clear queued standard output. Subsequent calls to `self.read_stdout()` will return
        ''.
        """
        self.read_stdout()

    def close(self):
        self.kill()

    @staticmethod
    def _kill_proc_tree(pid, including_parent=True):
        """Kill the process by pid, and any spawned child processes.
        What a dark metaphor.
        """
        try:
            parent = psutil.Process(pid)

            # Ever notice that this metaphor is very dark?
            children = parent.children(recursive=False)

            # Get the parent first to prevent respawning
            if including_parent:
                parent.kill()
                parent.wait(5)

            for child in children:
                ShellBackend._kill_proc_tree(child.pid)
        except psutil.NoSuchProcess:
            pass


class DotNetDevice(Device):
    """Base class for .NET library wrappers based on pythonnet.

    To implement a DotNetDevice subclass::

        import labbench as lb

        class MyLibraryWrapper(lb.DotNetDevice, libary=<imported python module colocated with dll>, dll_name='mylibrary.dll')
            ...

    When a DotNetDevice is instantiated, it looks to load the dll from the location of the python module
    and dll_name.

    Attributes:
        - `backend` is None after open and is available for replacement by the subclass
    """

    # these can only be set as arguments to a subclass definition
    library = value.any(None, allow_none=True, sets=False)  # Must be a module
    dll_name = value.str(None, allow_none=True, sets=False)

    _dlls = {}

    def open(self):
        """dynamically import a .net CLR as a python module at self.dll"""
        library = type(self).library.default
        dll_name = type(self).dll_name.default
        dll_path = Path(library.__path__[0]) / dll_name

        try:
            # static linters really don't like this, since it's created dynamically
            import clr
        except ImportError:
            raise ImportError("pythonnet module is required to use dotnet drivers")

        # base dotnet libraries needed to identify what we're working with
        clr.setPreload(False)
        clr.AddReference("System.Reflection")

        # more frustration for static linters
        import System
        from System.Reflection import Assembly

        try:
            contents = importlib.util.find_spec(library.__package__).loader.get_data(
                str(dll_path)
            )
        except BaseException:
            with open(dll_path, "rb") as f:
                contents = f.read()

        # binary file contents
        contents = importlib.util.find_spec(library.__package__).loader.get_data(
            str(dll_path)
        )

        # dump that into dotnet
        Assembly.Load(System.Array[System.Byte](contents))

        # do the actual import
        self.dll = importlib.import_module(dll_path.stem)


@property_.message_keying(write_fmt="{key} {value}", write_func="write")
class LabviewSocketInterface(Device):
    """Base class demonstrating simple sockets-based control of a LabView VI.

    Keyed get/set with lb.property are implemented by simple ' command value'.
    Subclasses can therefore implement support for commands in
    specific labview VI similar to VISA commands by
    assigning the commands implemented in the corresponding labview VI.

    Attributes:
        - backend: connection object mapping {'rx': rxsock, 'tx': txsock}
    """

    resource = value.NetworkAddress(
        "127.0.0.1", accept_port=False, help="LabView VI host address"
    )
    tx_port = value.int(61551, help="TX port to send to the LabView VI")
    rx_port = value.int(61552, help="TX port to send to the LabView VI")
    delay = value.float(1, help="time to wait after each property trait write or query")
    timeout = value.float(2, help="maximum wait replies before raising TimeoutError")
    rx_buffer_size = value.int(1024, min=1)

    def open(self):
        self.backend = dict(
            tx=socket.socket(socket.AF_INET, socket.SOCK_DGRAM),
            rx=socket.socket(socket.AF_INET, socket.SOCK_DGRAM),
        )

        self.backend["rx"].bind((self.resource, self.rx_port))
        self.backend["rx"].settimeout(self.timeout)
        self.clear()

    def close(self):
        for sock in list(self.backend.values()):
            try:
                sock.shutdown(socket.SHUT_RDWR)
                sock.close()
            except BaseException:
                self._logger.error("could not close socket ", repr(sock))

    def write(self, msg):
        """Send a string over the tx socket."""
        self._logger.debug(f"write {repr(msg)}")
        self.backend["tx"].sendto(msg, (self.resource, self.tx_port))
        util.sleep(self.delay)

    def read(self, convert_func=None):
        """Receive from the rx socket until `self.rx_buffer_size` samples
        are received or timeout happens after `self.timeout` seconds.

        Optionally, apply the conversion function to the value after
        it is received.
        """
        rx, addr = self.backend["rx"].recvfrom(self.rx_buffer_size)
        if addr is None:
            raise Exception("received no data")
        rx_disp = rx[: min(80, len(rx))] + ("..." if len(rx) > 80 else "")
        self._logger.debug(f"read {repr(rx_disp)}")

        key, value = rx.rsplit(" ", 1)
        key = key.split(":", 1)[1].lstrip()
        if convert_func is not None:
            value = convert_func(value)
        return {key: value}

    def clear(self):
        """Clear any data present in the read socket buffer."""
        while True:
            inputready, o, e = select.select([self.backend["rx"]], [], [], 0.0)
            if len(inputready) == 0:
                break
            for s in inputready:
                try:
                    s.recv(1)
                except BaseException:
                    continue


class SerialDevice(Device):
    """Base class for wrappers that communicate via pyserial.

    This implementation is very sparse because there is in general
    no messaging string format for serial devices.

    Attributes:
        - backend (serial.Serial): control object, after open
    """

    # Connection value traits
    resource = value.str(help="platform-dependent serial port address")
    timeout = value.float(
        2, min=0, help="Max time to wait for a connection before raising TimeoutError."
    )
    write_termination = value.bytes(
        b"\n", help="Termination character to send after a write."
    )
    baud_rate: int = value.int(
        9600, min=1, help="Data rate of the physical serial connection."
    )
    parity = value.bytes(b"N", help="Parity in the physical serial connection.")
    stopbits = value.float(
        1, min=1, max=2, step=0.5, help="Number of stop bits, one of `[1, 1.5, or 2.]`."
    )
    xonxoff = value.bool(False, help="`True` to enable software flow control.")
    rtscts = value.bool(False, help="`True` to enable hardware (RTS/CTS) flow control.")
    dsrdtr = value.bool(False, help="`True` to enable hardware (DSR/DTR) flow control.")

    # Overload methods as needed to implement the Device object protocol
    def open(self):
        """Connect to the serial device with the VISA resource string defined
        in self.resource
        """
        keys = "timeout", "parity", "stopbits", "xonxoff", "rtscts", "dsrdtr"
        params = dict([(k, getattr(self, k)) for k in keys])
        self.backend = serial.Serial(self.resource, self.baud_rate, **params)
        self._logger.debug(f"{repr(self)} connected")

    def close(self):
        """Disconnect the serial instrument"""
        self.backend.close()
        self._logger.debug(f"{repr(self)} closed")

    @classmethod
    def from_hwid(cls, hwid=None, *args, **connection_params):
        """Instantiate a new SerialDevice from a windows `hwid' string instead
        of a comport resource. A hwid string in windows might look something
        like:

        r'PCI\\VEN_8086&DEV_9D3D&SUBSYS_06DC1028&REV_21\\3&11583659&1&B3'
        """

        usb_map = cls._map_serial_hwid_to_port()
        if hwid not in usb_map:
            raise Exception(f"Cannot find serial port with hwid {repr(hwid)}")
        return cls(usb_map[hwid], *args, **connection_params)

    @staticmethod
    def list_ports(hwid=None):
        """List USB serial devices on the computer

        Returns:
            list of port resource information
        """
        from serial.tools import list_ports

        ports = [
            (port.device, {"hwid": port.hwid, "description": port.description})
            for port in list_ports.comports()
        ]
        ports = OrderedDict(ports)

        if hwid is not None:
            ports = [
                (port, meta) for port, meta in list(ports.items()) if meta["id"] == hwid
            ]

        return dict(ports)

    @staticmethod
    def _map_serial_hwid_to_label():
        """Map of the comports and their names.

        Returns:
            mapping {<comport name>: <comport ID>}
        """
        from serial.tools import list_ports

        return OrderedDict([(port[2], port[1]) for port in list_ports.comports()])

    @staticmethod
    def _map_serial_hwid_to_port():
        """Map of the comports and their names.

        Returns:
            mapping {<comport name>: <comport ID>}
        """
        from serial.tools import list_ports

        return OrderedDict([(port[2], port[0]) for port in list_ports.comports()])


class SerialLoggingDevice(SerialDevice):
    """Manage connection, acquisition, and data retreival on a single GPS device.
    The goal is to make GPS devices controllable somewhat like instruments:
    maintaining their own threads, and blocking during setup or stop
    command execution.

    Listener objects must implement an attach method with one argument
    consisting of the queue that the device manager uses to push data
    from the serial port.
    """

    poll_rate = value.float(
        0.1, min=0, help="Data retreival rate from the device (in seconds)"
    )
    data_format = value.bytes(b"", help="Data format metadata")
    stop_timeout = value.float(
        0.5, min=0, help="delay after `stop` before terminating run thread"
    )
    max_queue_size = value.int(
        100000, min=1, help="bytes to allocate in the data retreival buffer"
    )

    def configure(self):
        """This is called at the beginning of the logging thread that runs
        on a call to `start`.

        This is a stub that does nothing --- it should be implemented by a
        subclass for a specific serial logger device.
        """
        self._logger.debug(
            f"{repr(self)}: no device-specific configuration implemented"
        )

    def start(self):
        """Start a background thread that acquires log data into a queue.

        Returns:
            None
        """

        def accumulate():
            timeout, self.backend.timeout = self.backend.timeout, 0
            q = self._stdout
            stop_event = self._stop
            self._logger.debug(f"{repr(self)}: configuring log acquisition")
            self.configure()
            self._logger.debug(f"{repr(self)}: starting log acquisition")
            try:
                while stop_event.wait(self.poll_rate) is not True:
                    q.put(self.backend.read(10 * self.baud_rate * self.poll_rate))
            except serial.SerialException as e:
                self._stop.set()
                self.close()
                raise e
            finally:
                self._logger.debug(f"{repr(self)} ending log acquisition")
                try:
                    self.backend.timeout = timeout
                except BaseException:
                    pass

        if self.running():
            raise Exception("already running")

        self._stdout = Queue()
        self._stop = Event()
        Thread(target=accumulate).start()

    def stop(self):
        """Stops the logger acquisition if it is running. Returns silently otherwise.

        Returns:

            None
        """
        try:
            self._stop.set()
        except BaseException:
            pass

    def running(self):
        """Check whether the logger is running.

        Returns:
            `True` if the logger is running
        """
        return hasattr(self, "_stop") and not self._stop.is_set()

    def fetch(self):
        """Retrieve and return any log data in the buffer.

        Returns:

            any bytes in the buffer
        """
        ret = b""
        try:
            while True:
                ret += self._stdout.get_nowait()
        except Empty:
            pass
        return ret

    def clear(self):
        """Throw away any log data in the buffer."""
        self.fetch()

    def close(self):
        self.stop()


class TelnetDevice(Device):
    """A general base class for communication devices via telnet.
    Unlike (for example) VISA instruments, there is no
    standardized command format like SCPI. The implementation is
    therefore limited to open and close, which open
    or close a pyserial connection object: the `backend` attribute.
    Subclasses can read or write with the backend attribute like they
    would any other telnetlib instance.

    A TelnetDevice `resource` string is an IP address. The port is specified
    by `port`. These can be set when you instantiate the TelnetDevice
    or by setting them afterward in `value traits`.

    Subclassed devices that need property trait descriptors will need
    to implement get_key and set_key methods to implement
    the property trait set and get operations (as appropriate).
    """

    # Connection value traits
    resource = value.NetworkAddress("127.0.0.1:23", help="server host address")
    timeout = value.float(2, min=0, label="s", help="connection timeout")

    def open(self):
        """Open a telnet connection to the host defined
        by the string in self.resource
        """
        host, *port = self.resource.split(":")

        if len(port) > 0:
            port = int(port[0])
        else:
            port = 23

        self.backend = telnetlib.Telnet(self.resource, port=port, timeout=self.timeout)

    def close(self):
        """Disconnect the telnet connection"""
        self.backend.close()


@property_.visa_keying(
    query_fmt="{key}?", write_fmt="{key} {value}", remap={True: "ON", False: "OFF"}
)
class VISADevice(Device):
    r"""base class for VISA device wrappers with pyvisa.

    Examples:

        Autodetect a list of valid `resource` strings on the host::

            print(visa_list_resources())

        Fetch the instrument identity string::

            with VISADevice('USB0::0x2A8D::0x1E01::SG56360004::INSTR') as instr:
                print(inst.identity)

        Write ':FETCH?' to the instrument, read an expected ASCII CSV response,
        and return it as a pandas DataFrame::

            with VISADevice('USB0::0x2A8D::0x1E01::SG56360004::INSTR') as instr:
                print(inst.query_ascii_values(':FETCH?'))

    See also:
    .. _installing a proprietary OS service for VISA:
        https://pyvisa.readthedocs.io/en/latest/faq/getting_nivisa.html#faq-getting-nivisa
    .. _resource strings and basic configuration:
        https://pyvisa.readthedocs.io/en/latest/introduction/communication.html#getting-the-instrument-configuration-right

    Attributes:
        backend (pyvisa.Resource): instance of a pyvisa instrument object (when open)

    """

    # Settings
    read_termination = value.str(
        "\n", cache=True, help="end of line string to expect in query replies"
    )

    write_termination = value.str(
        "\n", cache=True, help="end of line string to send after writes"
    )

    open_timeout = value.float(
        None,
        cache=True,
        allow_none=True,
        help="timeout for opening a connection to the instrument",
        label="s",
    )

    identity_pattern = value.str(
        None,
        allow_none=True,
        cache=True,
        help="identity regex pattern to match for automatic connection",
    )

    timeout = value.float(
        None, cache=True, allow_none=True, help="message response timeout", label="s"
    )

    # Common VISA properties
    identity = property_.str(
        key="*IDN",
        sets=False,
        cache=True,
        help="identity string reported by the instrument",
    )

    @property_.dict(sets=False)
    def status_byte(self):
        """instrument status decoded from '*STB?'"""
        code = int(self.query("*STB?"))

        return {
            "error queue not empty": bool(code & 0b00000100),
            "questionable state": bool(code & 0b00001000),
            "message available": bool(code & 0b00010000),
            "event status flag": bool(code & 0b00100000),
            "service request": bool(code & 0b01000000),
            "top level status summary": bool(code & 0b01000000),
            "operating": bool(code & 0b10000000),
        }

    _rm = "@py"
    _opc = False

    # Overload methods as needed to implement RemoteDevice
    def open(self):
        """opens the instrument.

        When managing device connection through a `with` context,
        this is called automatically and does not need
        to be invoked.
        """
        from pyvisa import constants

        self._opc = False

        kwargs = dict(
            read_termination=self.read_termination,
            write_termination=self.write_termination,
        )

        if self.resource not in ("", None):
            pass
        elif self.identity_pattern is not None:
            pattern = re.compile(self.identity_pattern, flags=re.IGNORECASE)

            identities = {
                res: idn
                for res, idn in visa_list_identities().items()
                if re.match(pattern, idn) is not None
            }

            if len(identities) == 0:
                msg = f'could not open VISA device {repr(type(self))}: resource not specified, and no devices matched the pattern "{self.identity_pattern}"'
                raise IOError(msg)
            elif len(identities) == 1:
                self._logger.debug(
                    f'resource identified with identity pattern match "{list(identities.values())[0]}"'
                )
                self.resource = list(identities.keys())[0]
            else:
                msg = f'resource ambiguity: {len(identities)} VISA resources matched the pattern "{self.identity_pattern}"'
                raise IOError(msg)
        else:
            raise ValueError(
                f"specify the resource or identity_pattern attributes to open {repr(self)} connection"
            )

        if self.timeout is not None:
            kwargs["timeout"] = int(self.timeout * 1000)
        if self.open_timeout is not None:
            kwargs["open_timeout"] = int(self.open_timeout * 1000)

        # print(repr(kwargs))
        rm = self._get_rm()
        self.backend = rm.open_resource(self.resource, **kwargs)

    def close(self):
        """closes the instrument.

        When managing device connection through a `with` context,
        this is called automatically and does not need
        to be invoked.
        """
        if not self.isopen or self.backend is None:
            return

        try:
            with contextlib.suppress(pyvisa.errors.VisaIOError):
                self._release_remote_control()
            with contextlib.suppress(pyvisa.Error):
                self.backend.clear()

        except BaseException as e:
            e = str(e)
            if len(e.strip()) > 0:
                # some emulated backends raise empty errors
                self._logger.warning("unhandled close error: " + e)

        finally:
            self.backend.close()

    def write(self, msg: str):
        """sends an SCPI message to the device.

        Wraps `self.backend.write`, and handles debug logging and adjustments
        when in overlap_and_block
        contexts as appropriate.

        Arguments:
            msg: the SCPI command to send

        Returns:
            None
        """
        if self._opc:
            msg = msg + ";*OPC"
        msg_out = repr(msg) if len(msg) < 1024 else f"({len(msg)} bytes)"
        self._logger.debug(f"write {msg_out}")
        self.backend.write(msg)

    def query(self, msg: str, timeout=None, remap: bool = False) -> str:
        """queries the device with an SCPI message and returns its reply.

        Handles debug logging and adjustments when in overlap_and_block
        contexts as appropriate.

        Arguments:
            msg: the SCPI message to send
        """
        if timeout is not None:
            _to, self.backend.timeout = self.backend.timeout, timeout

        msg_out = repr(msg) if len(msg) < 80 else f"({len(msg)} bytes)"
        self._logger.debug(f"query {msg_out}")

        try:
            ret = self.backend.query(msg)
        finally:
            if timeout is not None:
                self.backend.timeout = _to

        msg_out = repr(ret) if len(ret) < 80 else f"({len(ret)} bytes)"
        self._logger.debug(f"    → {msg_out}")

        if remap:
            return self._property_keying.message_map[ret]
        else:
            return ret

    def query_ascii_values(
        self, msg: str, type_, separator=",", container=list, delay=None, timeout=None
    ):
        # pre debug
        if timeout is not None:
            _to, self.backend.timeout = self.backend.timeout, timeout

        msg_out = repr(msg) if len(msg) < 80 else f"({len(msg)} bytes)"
        self._logger.debug(f"query_ascii_values {msg_out}")

        try:
            ret = self.backend.query_ascii_values(
                msg, type_, separator, container, delay
            )
        finally:
            if timeout is not None:
                self.backend.timeout = _to

        # post debug
        if len(ret) < 80 and len(repr(ret)) < 80:
            logmsg = repr(ret)
        elif hasattr(ret, "shape"):
            logmsg = f"({type(ret).__qualname__} with shape {ret.shape})"
        elif hasattr(ret, "__len__"):
            logmsg = f"({type(ret).__qualname__} with length {len(ret)})"
        else:
            logmsg = f"(iterable sequence of type {type(ret)})"

        self._logger.debug(f"      -> {logmsg}")

        return ret

    def wait(self):
        """sends '*WAI' to wait for all commands to complete before continuing"""
        self.write("*WAI")

    def preset(self):
        """sends '*RST' to reset the instrument to preset"""
        self.write("*RST")

    @contextlib.contextmanager
    def overlap_and_block(self, timeout=None, quiet=False):
        """context manager that sends '*OPC' on entry, and performs
        a blocking '*OPC?' query on exit.

        By convention, these SCPI commands give a hint to the instrument
        that commands sent inside this block may be executed concurrently.
        The context exit then blocks until all of the commands have
        completed.

        Example::

            with inst.overlap_and_block():
                inst.write('long running command 1')
                inst.write('long running command 2')

        Arguments:
            timeout: maximum time to wait for '*OPC?' reply, or None to use `self.backend.timeout`
            quiet: Suppress timeout exceptions if this evaluates as True

        Raises:
            TimeoutError: on '*OPC?' query timeout
        """
        self._opc = True
        yield
        self._opc = False
        self.query("*OPC?", timeout=timeout)

    class suppress_timeout(contextlib.suppress):
        """context manager that suppresses timeout exceptions on `write` or `query`.

        Example::

            with inst.suppress_timeout():
                inst.write('long command 1')
                inst.write('long command 2')

            If the command 1 raises an exception, command 2 will not execute
            the context block is complete, and the exception from command 1
            is swallowed.
        """

        def __exit__(self, exctype, excinst, exctb):
            EXC = pyvisa.errors.VisaIOError
            CODE = pyvisa.errors.StatusCode.error_timeout

            return exctype == EXC and excinst.error_code == CODE

    def _release_remote_control(self):
        # From instrument and pyvisa docs
        if not self._rm.endswith("@sim"):
            self.backend.visalib.viGpibControlREN(
                self.backend.session, pyvisa.constants.VI_GPIB_REN_ADDRESS_GTL
            )

    def _get_rm(self):
        backend_name = self._rm

        if backend_name in ("@ivi", "@ni"):
            is_ivi = True
            # compatibility layer for changes in pyvisa 1.12
            if "ivi" in pyvisa.highlevel.list_backends():
                backend_name = "@ivi"
            else:
                backend_name = "@ni"
        else:
            is_ivi = False

        try:
            rm = pyvisa.ResourceManager(backend_name)
        except OSError as e:
            if is_ivi:
                url = r"https://pyvisa.readthedocs.io/en/latest/faq/getting_nivisa.html#faq-getting-nivisa"
                msg = f"could not connect to resource manager - see {url}"
                e.args[0] += msg
            raise e

        return rm


def visa_list_resources(resourcemanager: str = None):
    """autodetects and returns a list of valid resource strings"""
    if resourcemanager is None:
        rm = VISADevice()._get_rm()
    else:
        rm = pyvisa.ResourceManager(resourcemanager)

    return rm.list_resources()


def visa_default_resource_manager(name=None):
    VISADevice._rm = name


@util.TTLCache(timeout=3)
@util.single_threaded_call_lock
def visa_list_identities(skip_interfaces=["ASRL"], **device_kws) -> Dict[str, str]:
    def make_test_device(res):
        device = VISADevice(res, open_timeout=0.25, **device_kws)
        device._logger = logging.getLogger()
        return device

    def check_idn(device: VISADevice):
        try:
            return device.identity
        except pyvisa.errors.VisaIOError:
            return None

    def keep_interface(name):
        for iface in skip_interfaces:
            if name.lower().startswith(iface.lower()):
                return False
        return True

    devices = {
        res: make_test_device(res)
        for res in visa_list_resources()
        if keep_interface(res)
    }

    with util.sequentially(*list(devices.values()), catch=True):
        calls = [
            util.Call(check_idn, device).rename(res)
            for res, device in devices.items()
            if device.isopen
        ]

        identities = util.sequentially(*calls, catch=True)

    return identities


@Device.concurrency.adopt(True)
class Win32ComDevice(Device):
    """Basic support for calling win32 COM APIs.

    a dedicated background thread. Set concurrency=True to decide whether
    this thread support wrapper is applied to the dispatched Win32Com object.
    """

    # The python wrappers for COM drivers still basically require that
    # threading is performed using the windows COM API. Compatibility with
    # the python GIL is not for the faint of heart. Threading support is
    # instead realized with util.ThreadSandbox, which ensures that all calls
    # to the dispatched COM object block until the previous calls are completed
    # from within.

    com_object = value.str(
        default="", sets=False, help="the win32com object string"
    )  # Must be a module

    def open(self):
        """Connect to the win32 com object"""
        import win32com
        import win32com.client

        def should_sandbox(obj):
            try:
                name = win32com.__name__
                return inspect.getmodule(obj).__name__.startswith(name)
            except AttributeError:
                return False

        def factory():
            from pythoncom import CoInitialize

            CoInitialize()
            return win32com.client.Dispatch(self.com_object)

        # Oddness for win32 threadsafety
        sys.coinit_flags = 0

        if self.com_object == "":
            raise Exception("value traits.com_object needs to be set")

        if self.concurrency:
            self.backend = util.ThreadSandbox(factory, should_sandbox)
        else:
            self.backend = win32com.client.Dispatch(self.com_object)



# This software was developed by employees of the National Institute of
# Standards and Technology (NIST), an agency of the Federal Government.
# Pursuant to title 17 United States Code Section 105, works of NIST employees
# are not subject to copyright protection in the United States and are
# considered to be in the public domain. Permission to freely use, copy,
# modify, and distribute this software and its documentation without fee is
# hereby granted, provided that this notice and disclaimer of warranty appears
# in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' WITHOUT ANY WARRANTY OF ANY KIND, EITHER
# EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, ANY WARRANTY
# THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY IMPLIED WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND FREEDOM FROM
# INFRINGEMENT, AND ANY WARRANTY THAT THE DOCUMENTATION WILL CONFORM TO THE
# SOFTWARE, OR ANY WARRANTY THAT THE SOFTWARE WILL BE ERROR FREE. IN NO EVENT
# SHALL NIST BE LIABLE FOR ANY DAMAGES, INCLUDING, BUT NOT LIMITED TO, DIRECT,
# INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM,
# OR IN ANY WAY CONNECTED WITH THIS SOFTWARE, WHETHER OR NOT BASED UPON
# WARRANTY, CONTRACT, TORT, OR OTHERWISE, WHETHER OR NOT INJURY WAS SUSTAINED
# BY PERSONS OR PROPERTY OR OTHERWISE, AND WHETHER OR NOT LOSS WAS SUSTAINED
# FROM, OR AROSE OUT OF THE RESULTS OF, OR USE OF, THE SOFTWARE OR SERVICES
# PROVIDED HEREUNDER. Distributions of NIST software should also include
# copyright and licensing statements of any third-party software that are
# legally bundled with the code in compliance with the conditions of those
# licenses.

from . import _traits


class message_keying(_traits.PropertyKeyingBase):
    """Device class decorator that implements automatic API that triggers API messages for labbench properties.

    Example usage:

    ```python
        import labbench as lb

        @lb.message_keying(query_fmt='{key}?', write_fmt='{key} {value}', query_func='get', write_func='set')
        class MyDevice(lb.Device):
            def set(self, set_msg: str):
                # do set
                pass

            def get(self, get_msg: str):
                # do get
                pass
    ```

    Decorated classes connect traits that are defined with the `key` keyword to trigger
    backend API calls based on the key. The implementation of the `set` and `get` methods
    in subclasses of MessagePropertyAdapter determines how the key is used to generate API calls.
    """

    def __init__(
        self, query_fmt=None, write_fmt=None, write_func=None, query_func=None, remap={}
    ):
        super().__init__()

        self.query_fmt = query_fmt
        self.write_fmt = write_fmt
        self.write_func = write_func
        self.query_func = query_func

        if len(remap) == 0:
            self.value_map = {}
            self.message_map = {}
            return

        # ensure str type for messages; keys can be arbitrary python type
        if not all(isinstance(v, __builtins__["str"]) for v in remap.values()):
            raise TypeError("all values in remap dict must have type str")

        self.value_map = remap

        # create the reverse mapping
        self.message_map = __builtins__["dict"](zip(remap.values(), remap.keys()))

        # and ensure all values are unique
        if len(self.message_map) != len(self.value_map):
            raise ValueError("'remap' has duplicate values")

    def get(self, device: _traits.HasTraits, scpi_key: str, trait=None):
        """queries a parameter named `scpi_key` by sending an SCPI message string.

        The command message string is formatted as f'{scpi_key}?'.
        This is automatically called in wrapper objects on accesses to property traits that
        defined with 'key=' (which then also cast to a pythonic type).

        Arguments:
            key (str): the name of the parameter to set
            name (str, None): name of the trait setting the key (or None to indicate no trait) (ignored)

        Returns:
            response (str)
        """
        if self.query_fmt is None:
            raise ValueError("query_fmt needs to be set for key get operations")
        if self.query_func is None:
            raise ValueError("query_func needs to be set for key get operations")
        query_func = getattr(device, self.query_func)
        value_str = query_func(self.query_fmt.format(key=scpi_key)).rstrip()
        return self.message_map.get(value_str, value_str)

    def set(self, device: _traits.HasTraits, scpi_key: str, value, trait=None):
        """writes an SCPI message to set a parameter with a name key
        to `value`.

        The command message string is formatted as f'{scpi_key} {value}'. This
        This is automatically called on assignment to property traits that
        are defined with 'key='.

        Arguments:
            scpi_key (str): the name of the parameter to set
            value (str): value to assign
            name (str, None): name of the trait setting the key (or None to indicate no trait) (ignored)
        """
        if self.write_fmt is None:
            raise ValueError("write_fmt needs to be set for key set operations")
        if self.write_func is None:
            raise ValueError("write_func needs to be set for key set operations")

        value_str = self.value_map.get(value, value)
        write_func = getattr(device, self.write_func)
        write_func(self.write_fmt.format(key=scpi_key, value=value_str))


class visa_keying(message_keying):
    """Device class decorator that automates SCPI command string interactions for labbench properties.

    Example usage:

    ```python
        import labbench as lb

        @lb.property.visa_keying(query_fmt='{key}?', write_fmt='{key} {value}')
        class MyDevice(lb.VISADevice):
            pass
    ```

    This causes access to property traits defined with 'key=' to interact with the
    VISA instrument. By default, messages in VISADevice objects trigger queries
    with the `'{key}?'` format, and writes formatted as f'{key} {value}'.
    """

    def __init__(
        self,
        query_fmt="{key}?",
        write_fmt="{key} {value}",
        write_func="write",
        query_func="query",
        remap={},
    ):
        super().__init__(
            query_fmt=query_fmt,
            write_fmt=write_fmt,
            remap=remap,
            write_func="write",
            query_func="query",
        )


class bool(_traits.Bool):
    pass


class float(_traits.Float):
    pass


class int(_traits.Int):
    pass


class complex(_traits.Complex):
    pass


class str(_traits.Unicode):
    pass


class bytes(_traits.Bytes):
    pass


class list(_traits.List):
    pass


class tuple(_traits.Tuple):
    pass


class dict(_traits.Dict):
    pass


class Path(_traits.Path):
    pass


# class DataFrame(_traits.PandasDataFrame):
#     pass


# class Series(_traits.PandasSeries):
#     pass


# class ndarray(_traits.NumpyArray):
#     pass


class NetworkAddress(_traits.NetworkAddress):
    pass


# mutate these traits into the right role
_traits.subclass_namespace_traits(
    locals(), role=_traits.Trait.ROLE_PROPERTY, omit_trait_attrs=["default", "func"]
)



import typing
from . import util as util
from _typeshed import Incomplete
from collections.abc import Generator
from pathlib import Path
from typing import Any, Union

pd: Incomplete
Undefined: Incomplete
T = typing.TypeVar("T")

class ThisType(typing.Generic[T]): ...

class HasTraitsMeta(type):
    __cls_namespace__: Incomplete

    @classmethod
    def __prepare__(cls, names, bases, **kws): ...

class Trait:
    def __init__(
        default: str = None,
        key=None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
    ): ...
    ROLE_VALUE: str
    ROLE_PROPERTY: str
    ROLE_DATARETURN: str
    ROLE_UNSET: str
    type: Incomplete
    role = ROLE_UNSET
    default: ThisType
    key: Undefined
    func: typing.Callable
    help: str
    label: str
    sets: bool
    gets: bool
    cache: bool
    only: tuple
    allow_none: bool
    kws: Incomplete
    metadata: Incomplete

    @classmethod
    def __init_subclass__(cls, type=...) -> None: ...
    def copy(self, new_type: Incomplete | None = ..., **update_kws): ...
    __objclass__: Incomplete
    name: Incomplete

    def __set_name__(self, owner_cls, name) -> None: ...
    def __init_owner_subclass__(self, owner_cls) -> None: ...
    def __init_owner_instance__(self, owner) -> None: ...
    def __set__(self, owner: HasTraits, value): ...
    def __get__(self, owner, owner_cls: Incomplete | None = ...): ...
    def __cast_get__(self, owner, value, strict: bool = ...): ...
    def to_pythonic(self, value): ...
    def from_pythonic(self, value): ...
    def validate(self, value, owner: Incomplete | None = ...): ...
    def contains(self, iterable, value): ...
    def __call__(self, func=..., **kwargs): ...
    def doc(self, as_argument: bool = ..., anonymous: bool = ...): ...
    def doc_params(self, omit=...): ...
    def update(self, obj: Incomplete | None = ..., **attrs): ...
    def adopt(self, default=..., **trait_params): ...

def hold_trait_notifications(owner) -> Generator[None, None, None]: ...

class PropertyKeyingBase:
    def __new__(cls, *args, **kws): ...
    def __call__(self, owner_cls): ...
    def get(self, trait_owner, key, trait: Incomplete | None = ...) -> None: ...
    def set(self, trait_owner, key, value, trait: Incomplete | None = ...) -> None: ...

class HasTraits(metaclass=HasTraitsMeta):
    __notify_list__: Incomplete
    __cls_namespace__: Incomplete
    __cache__: Incomplete

    def __init__(self, **values) -> None: ...
    def __init_subclass__(cls) -> None: ...
    def __notify__(self, name, value, type, cache) -> None: ...
    def __get_value__(self, name): ...
    def __set_value__(self, name, value) -> None: ...

def adjusted(
    trait: Union[Trait, str], default: Any = ..., **trait_params
) -> HasTraits: ...

class Any(Trait, type=None):
    def __init__(
        default: str = None,
        key: str = None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
    ): ...
    def validate(self, value, owner: Incomplete | None = ...): ...
    def to_pythonic(self, value): ...

def observe(obj, handler, name=..., type_=...) -> None: ...
def unobserve(obj, handler) -> None: ...
def find_trait_in_mro(cls): ...

class DependentTrait(Trait):
    def __init__(
        default: str = None,
        key=None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
    ): ...
    def __set_name__(self, owner_cls, name) -> None: ...
    @classmethod
    def derive(
        mixin_cls, template_trait, dependent_traits=..., *init_args, **init_kws
    ): ...

class RemappingCorrectionMixIn(DependentTrait):
    def __init__(
        default: str = None,
        key=None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
        mapping: str = None,
    ): ...
    mapping: Any
    EMPTY_STORE: Incomplete

    def __init_owner_instance__(self, owner) -> None: ...
    def lookup_cal(self, uncal, owner): ...
    def find_uncal(self, cal, owner): ...
    def set_mapping(
        self,
        series_or_uncal,
        cal: Incomplete | None = ...,
        owner: Incomplete | None = ...,
    ) -> None: ...
    def __get__(self, owner, owner_cls: Incomplete | None = ...): ...
    def __set__(self, owner, cal) -> None: ...

class TableCorrectionMixIn(RemappingCorrectionMixIn):
    def __init__(
        default: str = None,
        key=None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
        mapping: str = None,
        table_index_column: str = None,
    ): ...
    path_trait: Incomplete
    index_lookup_trait: Incomplete
    table_index_column: str

    def __init_owner_instance__(self, owner) -> None: ...
    def __get__(self, owner, owner_cls: Incomplete | None = ...): ...
    def __set__(self, owner, cal) -> None: ...

class TransformMixIn(DependentTrait):
    def __init__(
        default: str = None,
        key=None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
    ): ...
    def __init_owner_instance__(self, owner) -> None: ...
    def __owner_event__(self, msg) -> None: ...
    def __get__(self, owner, owner_cls: Incomplete | None = ...): ...
    def __set__(self, owner, value_request) -> None: ...

class BoundedNumber(Trait):
    def __init__(
        default: str = None,
        key=None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = True,
        min: str = None,
        max: str = None,
        path_trait: str = None,
        index_lookup_trait: str = None,
        table_index_column: str = None,
    ): ...
    default: ThisType
    allow_none: bool
    min: ThisType
    max: ThisType

    def validate(self, value, owner: Incomplete | None = ...): ...
    path_trait: Any
    index_lookup_trait: Any
    table_index_column: str

    def calibrate_from_table(
        self,
        path_trait,
        index_lookup_trait,
        *,
        table_index_column: str = ...,
        help: str = ...,
        label=...,
        allow_none: bool = ...,
    ): ...
    def calibrate_from_expression(
        self,
        trait_expression,
        help: str = ...,
        label: str = ...,
        allow_none: bool = ...,
    ): ...
    def transform(
        self,
        other_trait: Trait,
        forward: callable,
        reverse: callable,
        help: str = ...,
        allow_none: bool = ...,
    ): ...
    def __neg__(self): ...
    def __add__(self, other): ...
    __radd__ = __add__

    def __sub__(self, other): ...
    def __rsub__(self, other): ...
    def __mul__(self, other): ...
    __rmul__ = __mul__

    def __truediv__(self, other): ...
    def __rdiv__(self, other): ...

class NonScalar(Any):
    def __init__(
        default: str = None,
        key: str = None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
    ): ...
    def validate(self, value, owner: Incomplete | None = ...): ...

class Int(BoundedNumber, type=int):
    def __init__(
        default: str = None,
        key=None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = True,
        min: str = None,
        max: str = None,
        path_trait: str = None,
        index_lookup_trait: str = None,
        table_index_column: str = None,
    ): ...
    ...

class Float(BoundedNumber, type=float):
    def __init__(
        default: str = None,
        key=None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = True,
        min: str = None,
        max: str = None,
        path_trait: str = None,
        index_lookup_trait: str = None,
        table_index_column: str = None,
        step: str = None,
    ): ...
    step: ThisType

    def validate(self, value, owner: Incomplete | None = ...): ...

class Complex(Trait, type=complex):
    def __init__(
        default: str = None,
        key=None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
    ): ...
    allow_none: bool

class Bool(Trait, type=bool):
    def __init__(
        default: str = None,
        key=None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
    ): ...
    allow_none: bool

    def validate(self, value, owner: Incomplete | None = ...): ...

class String(Trait):
    def __init__(
        default: str = None,
        key=None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
        case: str = True,
    ): ...
    case: bool

    def contains(self, iterable, value): ...

class Unicode(String, type=str):
    def __init__(
        default: str = "",
        key=None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
        case: str = True,
    ): ...
    default: ThisType

    def validate(self, value, owner: Incomplete | None = ...): ...

class Bytes(String, type=bytes):
    def __init__(
        default: str = b"",
        key=None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
        case: str = True,
    ): ...
    default: ThisType

class Iterable(Trait):
    def __init__(
        default: str = None,
        key=None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
    ): ...
    def validate(self, value, owner: Incomplete | None = ...): ...

class Dict(Iterable, type=dict):
    def __init__(
        default: str = None,
        key=None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
    ): ...
    ...

class List(Iterable, type=list):
    def __init__(
        default: str = None,
        key=None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
    ): ...
    ...

class Tuple(Iterable, type=tuple):
    def __init__(
        default: str = None,
        key=None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = False,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
    ): ...
    sets: bool

class Path(Trait, type=Path):
    def __init__(
        default: str = None,
        key=None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
        must_exist: str = False,
    ): ...
    must_exist: bool

    def validate(self, value, owner: Incomplete | None = ...): ...

class NetworkAddress(Unicode):
    def __init__(
        default: str = "",
        key=None,
        func: str = None,
        help: str = "",
        label: str = "",
        sets: str = True,
        gets: str = True,
        cache: str = False,
        only: str = (),
        allow_none: str = False,
        case: str = True,
        accept_port: str = True,
    ): ...
    accept_port: bool

    def validate(self, value, owner: Incomplete | None = ...): ...

VALID_TRAIT_ROLES: Incomplete

def subclass_namespace_traits(namespace_dict, role, omit_trait_attrs) -> None: ...



"""
.. This software was developed by employees of the National Institute of
.. Standards and Technology (NIST), an agency of the Federal Government.
.. Pursuant to title 17 United States Code Section 105, works of NIST employees
.. are not subject to copyright protection in the United States and are
.. considered to be in the public domain. Permission to freely use, copy,
.. modify, and distribute this software and its documentation without fee is
.. hereby granted, provided that this notice and disclaimer of warranty appears
.. in all copies.
..
.. THE SOFTWARE IS PROVIDED 'AS IS' WITHOUT ANY WARRANTY OF ANY KIND, EITHER
.. EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, ANY WARRANTY
.. THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY IMPLIED WARRANTIES OF
.. MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND FREEDOM FROM
.. INFRINGEMENT, AND ANY WARRANTY THAT THE DOCUMENTATION WILL CONFORM TO THE
.. SOFTWARE, OR ANY WARRANTY THAT THE SOFTWARE WILL BE ERROR FREE. IN NO EVENT
.. SHALL NIST BE LIABLE FOR ANY DAMAGES, INCLUDING, BUT NOT LIMITED TO, DIRECT,
.. INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM,
.. OR IN ANY WAY CONNECTED WITH THIS SOFTWARE, WHETHER OR NOT BASED UPON
.. WARRANTY, CONTRACT, TORT, OR OTHERWISE, WHETHER OR NOT INJURY WAS SUSTAINED
.. BY PERSONS OR PROPERTY OR OTHERWISE, AND WHETHER OR NOT LOSS WAS SUSTAINED
.. FROM, OR AROSE OUT OF THE RESULTS OF, OR USE OF, THE SOFTWARE OR SERVICES
.. PROVIDED HEREUNDER. Distributions of NIST software should also include
.. copyright and licensing statements of any third-party software that are
.. legally bundled with the code in compliance with the conditions of those
.. licenses.
"""

from .util import (
    Call,
    _force_full_traceback,
    concurrently,
    logger,
    retry,
    sequentially,
    show_messages,
    sleep,
    stopwatch,
    timeout_iter,
    until_timeout,
)

_force_full_traceback(True)

from . import datareturn, property, util, value
from ._backends import (
    DotNetDevice,
    LabviewSocketInterface,
    SerialDevice,
    SerialLoggingDevice,
    ShellBackend,
    TelnetDevice,
    VISADevice,
    Win32ComDevice,
    visa_default_resource_manager,
    visa_list_identities,
    visa_list_resources,
)
from ._data import CSVLogger, HDFLogger, SQLiteLogger, read, read_relational
from ._device import Device, list_devices, trait_info
from ._host import Email
from ._rack import (
    Rack,
    Sequence,
    find_owned_rack_by_type,
    import_as_rack,
    rack_input_table,
    rack_kwargs_skip,
    rack_kwargs_template,
)
from ._serialize import dump_rack, load_rack
from ._traits import PropertyKeyingBase, Undefined, adjusted, observe, unobserve
from ._version import __version__

# scrub __module__ for cleaner repr() and doc
for _obj in dict(locals()).values():
    if getattr(_obj, "__module__", "").startswith("labbench."):
        _obj.__module__ = "labbench"
del _obj

_force_full_traceback(True)



import labbench as lb
from _typeshed import Incomplete

class PowerSensor(lb.VISADevice):
    def __init__(
        self,
        resource: str = "str",
        read_termination: str = "str",
        write_termination: str = "str",
        open_timeout: str = "NoneType",
        identity_pattern: str = "str",
        timeout: str = "NoneType",
    ): ...
    RATES: Incomplete
    initiate_continuous: Incomplete
    trigger_count: Incomplete
    measurement_rate: Incomplete
    sweep_aperture: Incomplete
    frequency: Incomplete

    def preset(self) -> None: ...
    def fetch(self): ...
    def trigger(self): ...

class SpectrumAnalyzer(lb.VISADevice):
    def __init__(
        self,
        resource: str = "str",
        read_termination: str = "str",
        write_termination: str = "str",
        open_timeout: str = "NoneType",
        identity_pattern: str = "str",
        timeout: str = "NoneType",
    ): ...
    center_frequency: Incomplete
    resolution_bandwidth: Incomplete

    def load_state(self, remote_filename: str): ...
    def fetch(self): ...
    def trigger(self): ...

class SignalGenerator(lb.VISADevice):
    def __init__(
        self,
        resource: str = "str",
        read_termination: str = "str",
        write_termination: str = "str",
        open_timeout: str = "NoneType",
        identity_pattern: str = "str",
        timeout: str = "NoneType",
    ): ...
    output_enabled: Incomplete
    center_frequency: Incomplete
    mode: Incomplete

    def trigger(self) -> None: ...



import logging
from . import _device as core
from _typeshed import Incomplete

class LogStreamBuffer:
    def __init__(self) -> None: ...
    def write(self, msg): ...
    def read(self): ...
    def flush(self) -> None: ...

class LogStderr(core.Device):
    def __init__(self, resource: str = "str"): ...
    log: str

    def open(self) -> None: ...
    def close(self) -> None: ...
    def write(self, what) -> None: ...
    def flush(self) -> None: ...

class Email(core.Device):
    def __init__(
        self,
        resource: str = "str",
        port: str = "int",
        sender: str = "str",
        recipients: str = "list",
        success_message: str = "str",
        failure_message: str = "str",
    ): ...
    resource: Incomplete
    port: Incomplete
    sender: Incomplete
    recipients: Incomplete
    success_message: Incomplete
    failure_message: Incomplete
    backend: Incomplete

    def open(self) -> None: ...
    def close(self) -> None: ...
    def send_summary(self): ...

class JSONFormatter(logging.Formatter):
    t0: Incomplete

    def __init__(self, *args, **kws) -> None: ...
    @staticmethod
    def json_serialize_dates(obj): ...
    def format(self, rec): ...

class Host(core.Device):
    def __init__(self, resource: str = "str", git_commit_in: str = "NoneType"): ...
    git_commit_in: Incomplete
    time_format: str
    backend: Incomplete

    def open(self) -> None: ...
    def close(self) -> None: ...
    def metadata(self): ...
    def time(self): ...
    def log(self): ...
    def git_commit_id(self): ...
    def git_remote_url(self): ...
    def hostname(self): ...
    def git_browse_url(self): ...
    def git_pending_changes(self): ...



import io
from . import _device as core, _host, util as util, value as value
from ._device import Device as Device
from ._rack import Owner as Owner, Rack as Rack
from ._traits import observe as observe
from _typeshed import Incomplete
from collections.abc import Generator
from pathlib import Path
from typing import Dict, List, Union

h5py: Incomplete
np: Incomplete
pd: Incomplete
sqlalchemy: Incomplete
feather: Incomplete
EMPTY: Incomplete
INSPECT_SKIP_FILES: Incomplete

class MungerBase(core.Device):
    def __init__(
        self,
        resource: str = "WindowsPath",
        text_relational_min: str = "int",
        force_relational: str = "list",
        relational_name_fmt: str = "str",
        nonscalar_file_type: str = "str",
        metadata_dirname: str = "str",
    ): ...
    resource: Incomplete
    text_relational_min: Incomplete
    force_relational: Incomplete
    relational_name_fmt: Incomplete
    nonscalar_file_type: Incomplete
    metadata_dirname: Incomplete

    def __call__(self, index, row): ...
    def save_metadata(self, name, key_func, **extra): ...
    def open(self) -> None: ...

class MungeToDirectory(MungerBase):
    def __init__(
        self,
        resource: str = "WindowsPath",
        text_relational_min: str = "int",
        force_relational: str = "list",
        relational_name_fmt: str = "str",
        nonscalar_file_type: str = "str",
        metadata_dirname: str = "str",
    ): ...
    ...

class TarFileIO(io.BytesIO):
    tarfile: Incomplete
    overwrite: bool
    name: Incomplete
    mode: Incomplete

    def __init__(
        self, open_tarfile, relname, mode: str = ..., overwrite: bool = ...
    ) -> None: ...
    def __del__(self) -> None: ...
    def close(self) -> None: ...

class MungeToTar(MungerBase):
    def __init__(
        self,
        resource: str = "WindowsPath",
        text_relational_min: str = "int",
        force_relational: str = "list",
        relational_name_fmt: str = "str",
        nonscalar_file_type: str = "str",
        metadata_dirname: str = "str",
    ): ...
    tarname: str
    tarfile: Incomplete

    def open(self) -> None: ...
    def close(self) -> None: ...

class Aggregator(util.Ownable):
    PERSISTENT_TRAIT_ROLES: Incomplete
    name_map: Incomplete
    trait_rules: Incomplete
    incoming_trait_auto: Incomplete
    incoming_trait_always: Incomplete
    incoming_rack_output: Incomplete
    incoming_rack_input: Incomplete
    metadata: Incomplete

    def __init__(self) -> None: ...
    def enable(self) -> None: ...
    def disable(self) -> None: ...
    def is_always_trait(self, device, attr): ...
    def get(self) -> None: ...
    def key(self, device_name, state_name): ...
    def set_device_labels(self, **mapping: Dict[Device, str]): ...
    def update_name_map(
        self, ownables, owner_prefix: Incomplete | None = ...
    ) -> None: ...
    def observe(
        self,
        devices,
        changes: bool = ...,
        always: Union[str, List[str]] = ...,
        never: Union[str, List[str]] = ...,
    ): ...
    def inspect_object_name(self, target, max_levels: int = ...): ...

class TabularLoggerBase(
    Owner, util.Ownable, entry_order=(_host.Email, MungerBase, _host.Host)
):
    INDEX_LABEL: str
    path: Incomplete
    last_row: int
    pending_output: Incomplete
    pending_input: Incomplete
    aggregator: Incomplete
    host: Incomplete
    munge: Incomplete

    def __init__(
        self,
        path: Path = ...,
        *,
        append: bool = ...,
        text_relational_min: int = ...,
        force_relational: List[str] = ...,
        nonscalar_file_type: str = ...,
        metadata_dirname: str = ...,
        tar: bool = ...,
        git_commit_in: str = ...,
    ) -> None: ...
    def __copy__(self): ...
    def __owner_init__(self, owner: Owner): ...
    def observe(
        self,
        devices,
        changes: bool = ...,
        always: Union[str, List[str]] = ...,
        never: Union[str, List[str]] = ...,
    ): ...
    def set_row_preprocessor(self, func): ...
    def new_row(self, *args, **kwargs) -> None: ...
    def write(self) -> None: ...
    def context(self, *args, **kws) -> Generator[Incomplete, None, None]: ...
    def clear(self) -> None: ...
    def set_relational_file_format(self, format: str): ...
    output_index: int

    def open(self): ...
    def close(self) -> None: ...

class CSVLogger(TabularLoggerBase):
    ROOT_FILE_NAME: str
    OUTPUT_FILE_NAME: str
    INPUT_FILE_NAME: str
    output_index: int
    tables: Incomplete
    nonscalar_file_type: str

    def open(self) -> None: ...

class MungeToHDF(Device):
    def __init__(self, resource: str = "WindowsPath", key_fmt: str = "str"): ...
    resource: Incomplete
    key_fmt: Incomplete
    backend: Incomplete

    def open(self) -> None: ...
    def close(self) -> None: ...
    def __call__(self, index, row): ...
    def save_metadata(self, name, key_func, **extra): ...

class HDFLogger(TabularLoggerBase):
    KEY_OUTPUT: str
    KEY_INPUT: str
    nonscalar_file_type: str
    munge: Incomplete

    def __init__(
        self,
        path: Path = ...,
        *,
        append: bool = ...,
        key_fmt: str = ...,
        git_commit_in: str = ...,
    ) -> None: ...
    df: Incomplete

    def open(self) -> None: ...
    def close(self) -> None: ...

class SQLiteLogger(TabularLoggerBase):
    INDEX_LABEL: str
    ROOT_FILE_NAME: str
    OUTPUT_TABLE_NAME: str
    inprogress: Incomplete
    committed: Incomplete
    output_index: int

    def open(self) -> None: ...
    def close(self) -> None: ...
    def key(self, name, attr): ...

def to_feather(data, path) -> None: ...
def read_sqlite(
    path,
    table_name=...,
    columns: Incomplete | None = ...,
    nrows: Incomplete | None = ...,
    index_col=...,
): ...
def read(
    path_or_buf: str,
    columns: List[str] = ...,
    nrows: int = ...,
    format: str = ...,
    **kws,
): ...

class MungeTarReader:
    tarnames: Incomplete
    tarfile: Incomplete

    def __init__(self, path, tarname: str = ...) -> None: ...
    def __call__(self, key, *args, **kws): ...

class MungeDirectoryReader:
    path: Incomplete

    def __init__(self, path) -> None: ...
    def __call__(self, key, *args, **kws): ...

class MungeReader:
    def __new__(cls, path): ...

def read_relational(
    path: Union[str, Path],
    expand_col: str,
    root_cols: Union[List[str], None] = ...,
    target_cols: Union[List[str], None] = ...,
    root_nrows: Union[int, None] = ...,
    root_format: str = ...,
    prepend_column_name: bool = ...,
): ...



# This software was developed by employees of the National Institute of
# Standards and Technology (NIST), an agency of the Federal Government.
# Pursuant to title 17 United States Code Section 105, works of NIST employees
# are not subject to copyright protection in the United States and are
# considered to be in the public domain. Permission to freely use, copy,
# modify, and distribute this software and its documentation without fee is
# hereby granted, provided that this notice and disclaimer of warranty appears
# in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' WITHOUT ANY WARRANTY OF ANY KIND, EITHER
# EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, ANY WARRANTY
# THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY IMPLIED WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND FREEDOM FROM
# INFRINGEMENT, AND ANY WARRANTY THAT THE DOCUMENTATION WILL CONFORM TO THE
# SOFTWARE, OR ANY WARRANTY THAT THE SOFTWARE WILL BE ERROR FREE. IN NO EVENT
# SHALL NIST BE LIABLE FOR ANY DAMAGES, INCLUDING, BUT NOT LIMITED TO, DIRECT,
# INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM,
# OR IN ANY WAY CONNECTED WITH THIS SOFTWARE, WHETHER OR NOT BASED UPON
# WARRANTY, CONTRACT, TORT, OR OTHERWISE, WHETHER OR NOT INJURY WAS SUSTAINED
# BY PERSONS OR PROPERTY OR OTHERWISE, AND WHETHER OR NOT LOSS WAS SUSTAINED
# FROM, OR AROSE OUT OF THE RESULTS OF, OR USE OF, THE SOFTWARE OR SERVICES
# PROVIDED HEREUNDER. Distributions of NIST software should also include
# copyright and licensing statements of any third-party software that are
# legally bundled with the code in compliance with the conditions of those
# licenses.

from . import _device as core

from ._backends import VISADevice
from ._host import Host
from .util import show_messages
from ._rack import Rack
from ._traits import observe

import logging
import time
from io import StringIO
import numbers

import pandas as pd

import ipywidgets as widgets
from IPython.display import display

skip_traits = {VISADevice: ["identity"], Host: ["log"], core.Device: ["isopen"]}


def trait_table(device):
    """Generate a formatted html table widget which updates with recently-observed properties
    in a device.

    Arguments:
        device: the device to monitor, an deviceance of :class:`labbench.Device` (or one of its subclasses)
    Returns:
        `ipywidgdets.HBox` containing one `ipywidgets.HTML` widget
    """

    _df = pd.DataFrame([], columns=["value"])

    TABLE_STYLES = [
        {"selector": ".col_heading, .blank", "props": [("display", "none;")]}
    ]

    CAPTION_FMT = "<center><b>{}<b></center>"

    skip_attrs = []
    for cls, skip in skip_traits.items():
        if isinstance(device, cls):
            skip_attrs += skip

    html = widgets.HTML()

    def on_change(change):
        obj, name, value = change["owner"], change["name"], change["new"]

        if name != "isopen":
            print(name, value)

        # if name == 'isopen':
        #     if value:
        #         html.layout.visibility = 'visible'
        #     else:
        #         html.layout.visibility = 'hidden'

        if name in skip_attrs:
            return

        if hasattr(obj, "isopen") and name != "isopen" and not obj.isopen:
            if name in _df.index:
                _df.drop(name, inplace=True)
            return
        else:
            print(name)
        label = obj._traits[name].label
        _df.loc[name] = (str(value) + " " + str("" if label is None else label),)
        _df.sort_index(inplace=True)
        caption = CAPTION_FMT.format(obj._owned_name or repr(obj)).replace(",", "<br>")
        html.value = (
            _df.style.set_caption(caption)
            .set_table_attributes('class="table"')
            .set_table_styles(TABLE_STYLES)
            .render()
        )

    observe(device, on_change)

    return widgets.HBox([html])


class TextareaLogHandler(logging.StreamHandler):
    log_format = "%(asctime)s.%(msecs).03d %(levelname)10s %(message)s"
    time_format = "%Y-%m-%d %H:%M:%S"
    max_buffer = 10000
    min_delay = 0.1

    def __init__(self, level=logging.DEBUG):
        self.stream = StringIO()
        super(TextareaLogHandler, self).__init__(self.stream)
        self.widget = widgets.Textarea(
            layout=widgets.Layout(width="100%", height="500px")
        )
        self.setFormatter(logging.Formatter(self.log_format, self.time_format))
        self.setLevel(level)
        self.last_time = None

    def emit(self, record):
        ret = super(TextareaLogHandler, self).emit(record)
        if self.last_time is None or time.time() - self.last_time > self.min_delay:
            self.last_time = time.time()
            newvalue = self.widget.value + self.stream.getvalue()
            if len(newvalue) > self.max_buffer:
                newvalue = newvalue[-self.max_buffer :]
            self.widget.value = newvalue
        return ret


class panel:
    """Show tables summarizing value traits and property traits in jupyter notebook.
    Only a single panel will be shown in a python kernel.

    Arguments:
        source: Either an integer indicating how far up the calling tree to
                search for Device instances, or a `labbench.Rack` instance.
        ncols: Maximum number of devices to show on each row
    """

    widget = None
    ncols = 2
    devices = {}
    children = []

    def __new__(cls, source=1, ncols=2):
        cls.ncols = ncols

        if isinstance(source, Rack):
            cls.devices = dict(
                [
                    (k, v)
                    for k, v in source._ownables.items()
                    if isinstance(v, core.Device)
                ]
            )
        elif isinstance(source, numbers.Number):
            cls.source = source + 1
            cls.devices = core.list_devices(cls.source)
        else:
            raise ValueError(
                f"source must be a Rack instance or int, but got {repr(source)}"
            )

        children = [
            trait_table(cls.devices[k])
            for k in sorted(cls.devices.keys())
            if isinstance(cls.devices[k], core.Device)
        ]

        if len(children) == 0:
            return cls

        hboxes = []
        while True:
            N = min(ncols, len(children))

            try:
                hboxes.append(widgets.HBox(children[:N]))

            # Sometimes stale source._contexts leads to AttributeError.
            # Delete them and try again
            except AttributeError:
                if hasattr(source, "_contexts"):
                    return cls(source=cls.source, ncols=cls.ncols)
                else:
                    raise
            if N < len(children):
                children = children[N:]
            else:
                break

        vbox = widgets.VBox(hboxes)

        show_messages("error")
        log_handler = TextareaLogHandler()
        logger = logging.getLogger("labbench")

        logger.addHandler(log_handler)

        cls.widget = widgets.Tab([vbox, log_handler.widget])

        cls.widget.set_title(0, "State")
        cls.widget.set_title(1, "Debug")

        display(cls.widget)

        return cls



# This software was developed by employees of the National Institute of
# Standards and Technology (NIST), an agency of the Federal Government.
# Pursuant to title 17 United Property Code Section 105, works of NIST employees
# are not subject to copyright protection in the United Property and are
# considered to be in the public domain. Permission to freely use, copy,
# modify, and distribute this software and its documentation without fee is
# hereby granted, provided that this notice and disclaimer of warranty appears
# in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' WITHOUT ANY WARRANTY OF ANY KIND, EITHER
# EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, ANY WARRANTY
# THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY IMPLIED WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND FREEDOM FROM
# INFRINGEMENT, AND ANY WARRANTY THAT THE DOCUMENTATION WILL CONFORM TO THE
# SOFTWARE, OR ANY WARRANTY THAT THE SOFTWARE WILL BE ERROR FREE. IN NO EVENT
# SHALL NIST BE LIABLE FOR ANY DAMAGES, INCLUDING, BUT NOT LIMITED TO, DIRECT,
# INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM,
# OR IN ANY WAY CONNECTED WITH THIS SOFTWARE, WHETHER OR NOT BASED UPON
# WARRANTY, CONTRACT, TORT, OR OTHERWISE, WHETHER OR NOT INJURY WAS SUSTAINED
# BY PERSONS OR Decorator OR OTHERWISE, AND WHETHER OR NOT LOSS WAS SUSTAINED
# FROM, OR AROSE OUT OF THE RESULTS OF, OR USE OF, THE SOFTWARE OR SERVICES
# PROVIDED HEREUNDER. Distributions of NIST software should also include
# copyright and licensing statements of any third-party software that are
# legally bundled with the code in compliance with the conditions of those
# licenses.

import unittest
import sys
import labbench as lb

lb._force_full_traceback(True)


class TestProperty:
    TestDevice = None

    def test_get_set(self):
        with self.TestDevice() as m:
            m.clear_counts()

            # remapped bool "get"
            self.assertEqual(m.int0, m.PYTHONIC_VALUE_DEFAULT["int0"])

            # "set"
            new_param = m.int0 + 1
            m.int0 = new_param
            self.assertEqual(m.int0, new_param)
            self.assertEqual(m.remote_values["int0"], new_param)

            # verify expected "get"
            self.assertEqual(m.bool0, m.PYTHONIC_VALUE_DEFAULT["bool0"])  # get 1

            # verify expected "set"
            new_flag = not m.bool0  # get 2
            m.bool0 = new_flag

            self.assertEqual(m.bool0, new_flag)  # get 3

            self.assertEqual(
                m.remote_values["bool0"], m.get_expected_remote_value("bool0", new_flag)
            )

            self.assertEqual(m._getter_counts["bool0"], 3)
            self.assertEqual(m._getter_counts["int0"], 3)


class MockBase(lb.Device):
    # dictionary of mapping dictionaries, keyed on class attribute name
    REMAP = {}

    # the default value for bool traits {trait_name: trait_value}
    PYTHONIC_VALUE_DEFAULT = {}

    # test values for setting the property traits {trait_name: trait_value}
    PYTHONIC_VALUE_UPDATE = {}

    _getter_counts = {}

    def open(self):
        self._getter_counts = {}
        self.remote_values = {}
        self._last = {}

        for name, value in self.PYTHONIC_VALUE_DEFAULT.items():
            self.remote_values[name] = self.get_expected_remote_value(name, value)

    def add_get_count(self, name):
        self._getter_counts.setdefault(name, 0)
        self._getter_counts[name] += 1

    def clear_counts(self):
        self._getter_counts = {}

    def get_get_count(self, name):
        return self._getter_counts.get(name, 0)

    def get_expected_remote_value(self, name, value):
        return self._traits[name].type(value)


class MockDirectProperty(MockBase):
    PYTHONIC_VALUE_DEFAULT = {
        "int0": 3,
        "bool0": False,
        "str0": "moose",
        "str1": "moose",
        "str2": "moose",
    }

    PYTHONIC_VALUE_UPDATE = {
        "int0": 4,
        "bool0": True,
        "str0": "hi",
        "str1": "hi",
        "str2": "hi",
    }

    @lb.property.int(min=0, max=10)
    def int0(self, value):
        self.remote_values["int0"] = value

    def int0(self):
        self.add_get_count("int0")
        return self.remote_values["int0"]

    @lb.property.bool()
    def bool0(self, value):
        self.remote_values["bool0"] = value

    def bool0(self):
        self.add_get_count("bool0")
        return self.remote_values["bool0"]

    @lb.property.str(key="str0", cache=True)
    def str0(self, value):
        self.remote_values["str0"] = value

    def str0(self):
        self.add_get_count("str0")
        return self.remote_values["str0"]

    @lb.property.str(key="str1", cache=True)
    def str1(self, value):
        self.remote_values["str1"] = value

    def str1(self):
        self.add_get_count("str1")
        return self.remote_values["str1"]

    @lb.property.str(key="str2", sets=False)
    def str2(self, value):
        self.remote_values["str2"] = value

    def str2(self):
        self.add_get_count("str2")
        return self.remote_values["str2"]


class MockPropertyAdapter(lb.MessagePropertyAdapter):
    def get(self, device, key, trait=None):
        device.add_get_count(key)
        v = self.remote_values[key]
        return self.value_map.get(v, v)

    def set(self, device, key, value, trait=None):
        value = self.message_map.get(value, value)
        device._last[key] = value
        device.remote_values[key] = self.message_map.get(value, value)


@MockPropertyAdapter(remap={True: "ON", False: "OFF", "python value": "device value"})
class MockKeyedAdapterProperty(MockBase):
    # configure the emulated behavior of the fake instrument
    PYTHONIC_VALUE_DEFAULT = {
        "int0": 3,
        "bool0": False,
        "str0": "moose",
        "str1": "moose",
        "str2": "moose",
    }

    PYTHONIC_VALUE_UPDATE = {
        "int0": 4,
        "bool0": True,
        "str0": "hi",
        "str1": "hi",
        "str2": "hi",
    }

    int0 = lb.property.int(key="int0", min=0, max=10)
    bool0 = lb.property.bool(key="bool0")
    str0 = lb.property.str(key="str0", cache=True)
    str1 = lb.property.str(key="str1", cache=False)
    str2 = lb.property.str(key="str2", sets=False)


class MockDataReturn(MockBase):
    @lb.datareturn.float()
    def fetch_float0(self, a, b, c):
        return a + b + c

    @lb.datareturn.float
    def fetch_float1(self, a, b, c):
        return a + b + c


class TestProperty:
    # set this in a subclass
    TestDevice = None

    def test_set_then_get(self):
        with self.TestDevice() as m:
            m.clear_counts()

            for trait_name, new_value in m.PYTHONIC_VALUE_UPDATE.items():
                msg = f'property "{trait_name}"'
                trait = m._traits[trait_name]

                if not (trait.sets and trait.gets):
                    # this test is only for traits that support both set and get
                    continue

                m.clear_counts()
                setattr(m, trait_name, new_value)

                # in cases of remap, ensure the stored value in the mock device
                # matches the expected
                self.assertEqual(
                    m.remote_values[trait_name],
                    m.get_expected_remote_value(trait_name, new_value),
                    msg=msg,
                )

                # validate the pythonic value
                self.assertEqual(
                    getattr(m, trait_name), new_value, msg=msg
                )  # +1 get (unless cached)

                # make sure there weren't any unecessary extra 'get' operations
                self.assertEqual(
                    m.get_get_count(trait_name), 0 if trait.cache else 1, msg=msg
                )

    def test_disabled_set(self):
        with self.TestDevice() as m:
            m.clear_counts()

            for trait_name, new_value in m.PYTHONIC_VALUE_UPDATE.items():
                trait = m._traits[trait_name]
                msg = f'property "{trait_name}"'

                if not trait.sets:
                    with self.assertRaises(AttributeError, msg=msg):
                        setattr(m, trait_name, new_value)

    def test_cache(self):
        with self.TestDevice() as m:
            m.clear_counts()

            for trait_name, new_value in m.PYTHONIC_VALUE_UPDATE.items():
                if m._traits[trait_name].cache:
                    msg = f'property "{trait_name}"'

                    # Device gets should only happen once if cache=True
                    m.clear_counts()
                    self.assertEqual(
                        getattr(m, trait_name),
                        m.PYTHONIC_VALUE_DEFAULT[trait_name],
                        msg=msg,
                    )
                    self.assertEqual(
                        getattr(m, trait_name),
                        m.PYTHONIC_VALUE_DEFAULT[trait_name],
                        msg=msg,
                    )
                    self.assertEqual(m.get_get_count(trait_name), 1, msg=msg)

    def test_get_default(self):
        """
        Ensure that values are returned, typing is correct, and
        """
        with self.TestDevice() as m:
            m.clear_counts()

            # remapped bool "get"
            for trait_name, value in m.PYTHONIC_VALUE_DEFAULT.items():
                msg = f'testing property "{trait_name}"'
                self.assertEqual(getattr(m, trait_name), value, msg=msg)
                self.assertEqual(m.get_get_count(trait_name), 1, msg=msg)


class TestDecoratedProperty(unittest.TestCase, TestProperty):
    TestDevice = MockDirectProperty


class TestKeyedProperty(unittest.TestCase, TestProperty):
    TestDevice = MockDirectProperty


class TestReturner(unittest.TestCase):
    def test_returner_type(self):
        with MockDataReturn() as m:
            self.assertEqual(m.fetch_float0(1, 2, 3), 6.0)
            self.assertEqual(m.fetch_float1(1, 2, 3), 6.0)


if __name__ == "__main__":
    lb.show_messages("debug")
    unittest.main()

    # with MockDecoratedProperty() as m:
    #     pass



from _typeshed import Incomplete
from collections.abc import Generator
from threading import ThreadError
from typing import Callable

import_t0: Incomplete
logger: Incomplete

class LabbenchDeprecationWarning(DeprecationWarning): ...

def show_messages(minimum_level, colors: bool = ...) -> None: ...

class Ownable:
    __objclass__: Incomplete
    def __init__(self) -> None: ...
    __name__: Incomplete
    def __set_name__(self, owner_cls, name) -> None: ...
    def __get__(self, owner, owner_cls: Incomplete | None = ...): ...
    def __owner_init__(self, owner) -> None: ...
    def __owner_subclass__(self, owner_cls): ...

class ConcurrentException(Exception):
    thread_exceptions: Incomplete

class OwnerThreadException(ThreadError): ...
class ThreadEndedByMaster(ThreadError): ...

def hide_in_traceback(func): ...

class _filtered_exc_info:
    lb_wrapped: Incomplete
    def __init__(self, wrapped) -> None: ...
    def __call__(self): ...

def copy_func(func, assigned=..., updated=...) -> callable: ...
def sleep(seconds, tick: float = ...) -> None: ...
def check_hanging_thread() -> None: ...
def retry(
    exception_or_exceptions,
    tries: int = ...,
    delay: int = ...,
    backoff: int = ...,
    exception_func=...,
): ...
def until_timeout(
    exception_or_exceptions,
    timeout,
    delay: int = ...,
    backoff: int = ...,
    exception_func=...,
): ...
def timeout_iter(duration) -> Generator[Incomplete, None, None]: ...
def kill_by_name(*names) -> None: ...
def hash_caller(call_depth: int = ...): ...
def stopwatch(desc: str = ..., threshold: float = ..., logger_level: str = ...): ...

class Call:
    func: Incomplete
    name: Incomplete
    args: Incomplete
    kws: Incomplete
    queue: Incomplete
    def __init__(self, func, *args, **kws) -> None: ...
    def rename(self, name): ...
    result: Incomplete
    traceback: Incomplete
    def __call__(self): ...
    def set_queue(self, queue) -> None: ...
    @classmethod
    def wrap_list_to_dict(cls, name_func_pairs): ...

class MultipleContexts:
    abort: bool
    __name__: str
    objs: Incomplete
    params: Incomplete
    call_handler: Incomplete
    exc: Incomplete
    def __init__(
        self, call_handler: Callable[[dict, list, dict], dict], params: dict, objs: list
    ) -> None: ...
    def enter(self, name, context) -> None: ...
    def __enter__(self) -> None: ...
    def __exit__(self, *exc) -> None: ...

def concurrently(*objs, **kws): ...
def sequentially(*objs, **kws): ...

class ThreadDelegate:
    def __init__(self, sandbox, obj, dir_, repr_) -> None: ...
    def __call__(self, *args, **kws): ...
    def __getattribute__(self, name): ...
    def __dir__(self): ...
    def __setattr__(self, name, value): ...

class ThreadSandbox:
    __repr_root__: str
    __dir_root__: Incomplete
    def __init__(
        self, factory, should_sandbox_func: Incomplete | None = ...
    ) -> None: ...
    def __getattr__(self, name): ...
    def __setattr__(self, name, value): ...
    def __del__(self) -> None: ...
    def __dir__(self): ...

class single_threaded_call_lock:
    def __new__(cls, func): ...
    def __call__(self, *args, **kws): ...

def lazy_import(name): ...

class TTLCache:
    timeout: Incomplete
    call_timestamp: Incomplete
    last_value: Incomplete
    def __init__(self, timeout) -> None: ...
    def __call__(self, func): ...

# Names in __all__ with no definition:
#   _force_full_traceback



from . import util as util
from ._rack import (
    Rack as Rack,
    import_as_rack as import_as_rack,
    update_parameter_dict as update_parameter_dict,
)
from _typeshed import Incomplete
from pathlib import Path

RACK_CONFIG_FILENAME: str
EMPTY: Incomplete
INDEX_COLUMN_NAME: str

def write_table_stub(rack: Rack, name: str, path: Path, with_defaults: bool = ...): ...
def dump_rack(
    rack: Rack,
    output_path: Path,
    sourcepath: Path,
    pythonpath: Path = ...,
    exist_ok: bool = ...,
    with_defaults: bool = ...,
    skip_tables: bool = ...,
): ...
def read_yaml_config(config_path: str): ...
def load_rack(output_path: str, defaults: dict = ..., apply: bool = ...) -> Rack: ...



from __future__ import annotations
import typing
from typing import Callable, TypeVar, Dict, Tuple, Type, List, Sequence, cast, get_type_hints, Concatenate, Any, Protocol
from typing_extensions import TypeVarTuple, Unpack, ParamSpec
import functools
import labbench as lb
from dataclasses import dataclass, make_dataclass

Ts = TypeVarTuple('Ts')
P = ParamSpec('P')

def whatever(*args, **kws):
    """docs"""
    pass

def method_factory(*args: Unpack[Ts]) -> Callable[[Unpack[Ts]]]:
    @functools.wraps(whatever)
    def myfunc(*a):
        return 7

    return myfunc

func = method_factory(lb.value.str.__init__, lb.value.bytes)


print(func.__annotations__)


# This software was developed by employees of the National Institute of
# Standards and Technology (NIST), an agency of the Federal Government.
# Pursuant to title 17 United States Code Section 105, works of NIST employees
# are not subject to copyright protection in the United States and are
# considered to be in the public domain. Permission to freely use, copy,
# modify, and distribute this software and its documentation without fee is
# hereby granted, provided that this notice and disclaimer of warranty appears
# in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' WITHOUT ANY WARRANTY OF ANY KIND, EITHER
# EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, ANY WARRANTY
# THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY IMPLIED WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND FREEDOM FROM
# INFRINGEMENT, AND ANY WARRANTY THAT THE DOCUMENTATION WILL CONFORM TO THE
# SOFTWARE, OR ANY WARRANTY THAT THE SOFTWARE WILL BE ERROR FREE. IN NO EVENT
# SHALL NIST BE LIABLE FOR ANY DAMAGES, INCLUDING, BUT NOT LIMITED TO, DIRECT,
# INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM,
# OR IN ANY WAY CONNECTED WITH THIS SOFTWARE, WHETHER OR NOT BASED UPON
# WARRANTY, CONTRACT, TORT, OR OTHERWISE, WHETHER OR NOT INJURY WAS SUSTAINED
# BY PERSONS OR PROPERTY OR OTHERWISE, AND WHETHER OR NOT LOSS WAS SUSTAINED
# FROM, OR AROSE OUT OF THE RESULTS OF, OR USE OF, THE SOFTWARE OR SERVICES
# PROVIDED HEREUNDER. Distributions of NIST software should also include
# copyright and licensing statements of any third-party software that are
# legally bundled with the code in compliance with the conditions of those
# licenses.

"""
This implementation is deeply intertwined with obscure details of the python object
model. Consider starting with a close read of the documentation and exploring
the objects in an interpreter instead of reverse-engineering this code.
"""

from functools import wraps
import inspect
import sys
import traceback
from warnings import warn
from typing import Union

from . import util
from . import property as property_
from . import value

from ._traits import (
    HasTraits,
    Undefined,
    BoundedNumber,
    observe,
    unobserve,
    hold_trait_notifications,
)

__all__ = ["Device", "list_devices", "property", "value", "trait_info"]


def trace_methods(cls, name, until_cls=None):
    """Look for a method called `name` in cls and all of its parent classes."""
    methods = []
    last_method = None

    for cls in cls.__mro__:
        try:
            this_method = getattr(cls, name)
        except AttributeError:
            continue
        if this_method != last_method:
            methods.append(this_method)
            last_method = this_method
        if cls is until_cls:
            break

    return methods


def list_devices(depth=1):
    """Look for Device instances, and their names, in the calling
    code context (depth == 1) or its callers (if depth in (2,3,...)).
    Checks locals() in that context first.
    If no Device instances are found there, search the first
    argument of the first function argument, in case this is
    a method in a class.
    """
    from inspect import currentframe

    f = frame = currentframe()
    for i in range(depth):
        f = f.f_back
    try:
        ret = {k: v for k, v in list(f.frame.f_locals.items()) if isinstance(v, Device)}

        # If the context is a function, look in its first argument,
        # in case it is a method. Search its class instance.
        if len(ret) == 0 and len(f.frame.f_code.co_varnames) > 0:
            obj = f.frame.f_locals[f.frame.f_code.co_varnames[0]]
            for k, v in obj.__dict__.items():
                if isinstance(v, Device):
                    ret[k] = v
    finally:
        del f, frame

    return ret


class DisconnectedBackend(object):
    """ "Null Backend" implementation to raises an exception with discriptive
    messages on attempts to use a backend before a Device is connected.
    """

    def __init__(self, dev):
        """dev may be a class or an object for error feedback"""
        if isinstance(dev, str):
            self.name = dev
        elif getattr(dev, "_owned_name", None) is not None:
            self.name = dev._owned_name
        else:
            self.name = f"{dev.__class__.__qualname__} instance"

    @util.hide_in_traceback
    def __getattr__(self, key):
        msg = f"open {self.name} first to access its backend"
        raise ConnectionError(msg)

    def __repr__(self):
        return "DisconnectedBackend()"

    def __copy__(self, memo=None):
        return DisconnectedBackend(self.name)

    str = __repr__
    __deepcopy__ = __copy__


def log_trait_activity(msg):
    """emit debug messages for trait values"""

    if msg["name"] == "isopen":
        return

    owner = msg["owner"]
    trait_name = msg["name"]

    label = ""
    if msg["type"] == "set":
        if owner._traits[trait_name].label:
            label = f"({owner._traits[trait_name].label})"
        value = repr(msg["new"])
        if len(value) > 180:
            value = f'<data of type {type(msg["new"]).__qualname__}>'
        owner._logger.debug(f'trait set: "{trait_name}" → {value} {label}'.rstrip())
    elif msg["type"] == "get":
        if owner._traits[trait_name].label:
            label = f"({owner._traits[trait_name].label})"
        value = repr(msg["new"])
        if len(value) > 180:
            value = f'<data of type {type(msg["new"]).__qualname__}>'
        owner._logger.debug(f'trait get: "{trait_name}" → {value} {label}'.rstrip())
    else:
        owner._logger.debug(f'unknown operation type "{msg["type"]}"')


class Device(HasTraits, util.Ownable):
    r"""base class for labbench device wrappers.

    Drivers that subclass `Device` share

    * standardized connection management via context blocks (the `with` statement)
    * hooks for automatic data logging and heads-up displays
    * API style consistency
    * bounds checking and casting for typed attributes

    .. note::
        This `Device` base class has convenience
        functions for device control, but no implementation.

        Some wrappers for particular APIs labbench Device subclasses:

            * VISADevice: pyvisa,
            * ShellBackend: binaries and scripts
            * Serial: pyserial
            * DotNetDevice: pythonnet

        (and others). If you are implementing a driver that uses one of
        these backends, inherit from the corresponding class above, not
        `Device`.

    """

    resource = value.str(allow_none=True, cache=True, help="device address or URI")
    concurrency = value.bool(
        True, sets=False, help="True if the device supports threading"
    )
    """ Container for property trait traits in a Device. Getting or setting property trait traits
        triggers live updates: communication with the device to get or set the
        value on the Device. Therefore, getting or setting property trait traits
        needs the device to be connected.

        To set a property trait value inside the device, use normal python assigment::

            device.parameter = value

        To get a property trait value from the device, you can also use it as a normal python variable::

            variable = device.parameter + 1
    """
    backend = DisconnectedBackend(None)
    """ .. this attribute is some reference to a controller for the device.
        it is to be set in `connect` and `disconnect` by the subclass that implements the backend.
    """

    # Backend classes may optionally overload these, and do not need to call the parents
    # defined here
    def open(self):
        """Backend implementations overload this to open a backend
        connection to the resource.
        """
        observe(self, log_trait_activity)

    def close(self):
        """Backend implementations must overload this to disconnect an
        existing connection to the resource encapsulated in the object.
        """
        self.backend = DisconnectedBackend(self)
        self.isopen
        unobserve(self, log_trait_activity)

    __children__ = {}

    @util.hide_in_traceback
    def __init__(self, resource=Undefined, **values):
        """Update default values with these arguments on instantiation."""

        if hasattr(self, "__imports__"):
            warn(
                "the use of __imports__ has been deprecated. switch to importing each backend-specific module in each method that uses it."
            )
            self.__imports__()

        # validate presence of required arguments
        inspect.signature(self.__init__).bind(resource, **values)

        if resource is not Undefined:
            values["resource"] = resource

        super().__init__()

        with hold_trait_notifications(self):
            for name, init_value in values.items():
                if init_value != self._traits[name].default:
                    setattr(self, name, init_value)

        util.Ownable.__init__(self)

        self.backend = DisconnectedBackend(self)

        # Instantiate property trait now. It needed to wait until this point, after values are fully
        # instantiated, in case property trait implementation depends on values
        setattr(self, "open", self.__open_wrapper__)
        setattr(self, "close", self.__close_wrapper__)

    @classmethod
    @util.hide_in_traceback
    def __init_subclass__(cls):
        super().__init_subclass__()
        cls.__update_signature__()

    @classmethod
    @util.hide_in_traceback
    def __update_signature__(cls):
        # Generate a signature for documentation and code autocomplete
        params = [
            inspect.Parameter("self", kind=inspect.Parameter.POSITIONAL_ONLY),
            inspect.Parameter(
                "resource",
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                default=cls.resource.default,
                annotation=cls.resource.type,
            ),
        ]

        settable_values = {
            name: cls._traits[name]
            for name in cls._value_attrs
            if cls._traits[name].sets
        }

        # generate and apply the sequence of call signature parameters
        params += [
            inspect.Parameter(
                name,
                kind=inspect.Parameter.KEYWORD_ONLY,
                default=trait.default,
                annotation=trait.type,
            )
            for name, trait in settable_values.items()
            if name != "resource"
        ]

        # we need a wrapper so that __init__ can be modified separately for each subclass
        cls.__init__ = util.copy_func(cls.__init__)
        cls.__init__.__signature__ = inspect.Signature(params)

        # generate the __init__ docstring
        value_docs = "".join((f"    {t.doc()}\n" for t in settable_values.values()))
        cls.__init__.__doc__ = f"\nArguments:\n{value_docs}"

    @util.hide_in_traceback
    @wraps(open)
    def __open_wrapper__(self):
        """A wrapper for the connect() method. It steps through the
        method resolution order of self.__class__ and invokes each open()
        method, starting with labbench.Device and working down
        """
        if self.isopen:
            self._logger.debug(f"attempt to open {self}, which is already open")
            return

        self.backend = None

        try:
            for opener in trace_methods(self.__class__, "open", Device)[::-1]:
                opener(self)
        except BaseException:
            self.backend = DisconnectedBackend(self)
            raise

        self._logger.debug("opened")

        # Force an update to self.isopen
        self.isopen

    @util.hide_in_traceback
    @wraps(close)
    def __close_wrapper__(self):
        """A wrapper for the close() method that runs
        cleanup() before calling close(). close()
        will still be called if there is an exception in cleanup().
        """
        # Try to run cleanup(), but make sure to run
        # close() even if it fails
        # with self._hold_notifications('isopen'):
        if not self.isopen:
            return

        methods = trace_methods(self.__class__, "close", Device)

        all_ex = []
        for closer in methods:
            try:
                closer(self)
            except BaseException:
                all_ex.append(sys.exc_info())

        try:
            # Print tracebacks for any suppressed exceptions
            for ex in all_ex[::-1]:
                # If ThreadEndedByMaster was raised, assume the error handling in
                # util.concurrently will print the error message
                if ex[0] is not util.ThreadEndedByMaster:
                    depth = len(tuple(traceback.walk_tb(ex[2])))
                    traceback.print_exception(*ex, limit=-(depth - 1))
                    sys.stderr.write("(Exception suppressed to continue close)\n\n")

            self.isopen

            self._logger.debug("closed")
        finally:
            if len(all_ex) > 0:
                ex = util.ConcurrentException(
                    f"multiple exceptions while closing {self}"
                )
                ex.thread_exceptions = all_ex
                raise ex

    @util.hide_in_traceback
    def __enter__(self):
        try:
            self.open()
            return self
        except BaseException as e:
            args = list(e.args)
            if len(args) > 0:
                args[0] = f"{repr(self)}: {args[0]}"
                e.args = tuple(args)
            raise e

    @util.hide_in_traceback
    def __exit__(self, type_, value, traceback):
        try:
            self.close()
        except BaseException as e:
            args = list(e.args)
            args[0] = "{}: {}".format(repr(self), str(args[0]))
            e.args = tuple(args)
            raise e

    # Object boilerplate
    def __del__(self):
        try:
            isopen = self.isopen
        except AttributeError:
            # the object failed to instantiate properly
            isopen = False

        if isopen:
            self.close()

    def __repr__(self):
        name = self.__class__.__qualname__
        if hasattr(self, "resource"):
            return f"{name}({repr(self.resource)})"
        else:
            # In case an exception has occurred before __init__
            return f"{name}()"

    @property_.bool()
    def isopen(self):
        """`True` if the backend is ready for use"""
        try:
            return DisconnectedBackend not in self.backend.__class__.__mro__
        except BaseException:
            # Run into this sometimes on reloading a module or ipython shell:
            # the namespace is gone. we just assume disconnected
            return False


Device.__init_subclass__()


def trait_info(device: Device, name: str) -> dict:
    """returns the keywords used to define the trait attribute named `name` in `device`"""

    trait = device._traits[name]
    info = dict(trait.kws)

    if isinstance(trait, BoundedNumber):
        info.update(
            min=trait._min(device),
            max=trait._max(device),
            step=trait.step,
        )

    return info
